# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/ControlSurface.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 32416 bytes
import logging
import traceback
from contextlib import contextmanager
from functools import partial, wraps
from itertools import chain

import Live
from ableton.v2.base import old_hasattr

from . import Defaults, Task
from .ControlElement import OptimizedOwnershipHandler
from .Dependency import inject
from .InputControlElement import (
    MIDI_CC_TYPE,
    MIDI_NOTE_TYPE,
    MIDI_PB_STATUS,
    MIDI_PB_TYPE,
    MIDI_SYSEX_TYPE,
    InputControlElement,
)
from .MessageScheduler import MessageScheduler
from .PhysicalDisplayElement import PhysicalDisplayElement
from .Profile import profile
from .SubjectSlot import SlotManager, Subject
from .Util import BooleanContext, const, find_if, first

logger = logging.getLogger(__name__)


def _scheduled_method(method):
    @wraps(method)
    def wrapper(self, *a, **k):
        def doit():
            return method(self, *a, **k)

        self.schedule_message(1, doit)

    return wrapper


CS_LIST_KEY = "control_surfaces"


def publish_control_surface(control_surface):
    get_control_surfaces().append(control_surface)


def get_control_surfaces():
    if isinstance(__builtins__, dict):
        if CS_LIST_KEY not in list(__builtins__.keys()):
            __builtins__[CS_LIST_KEY] = []
        return __builtins__[CS_LIST_KEY]
    if not old_hasattr(__builtins__, CS_LIST_KEY):
        setattr(__builtins__, CS_LIST_KEY, [])
    return getattr(__builtins__, CS_LIST_KEY)


class ControlSurface(Subject, SlotManager):
    received_midi = ()
    __subject_events__ = ("received_midi", "disconnect")

    def __init__(self, c_instance=None, *a, **k):
        (super().__init__)(*a, **k)
        self.canonical_parent = None
        publish_control_surface(self)
        self._c_instance = c_instance
        self.log_message("Initializing...")
        self._pad_translations = None
        self._suggested_input_port = ""
        self._suggested_output_port = ""
        self._components = []
        self._displays = []
        self.controls = []
        self._highlighting_session_component = None
        self._device_component = None
        self._forwarding_long_identifier_registry = {}
        self._forwarding_registry = {}
        self._is_sending_scheduled_messages = BooleanContext()
        self._remaining_scheduled_messages = []
        self._task_group = Task.TaskGroup(auto_kill=False)
        self._in_build_midi_map = BooleanContext()
        self._suppress_requests_counter = 0
        self._rebuild_requests_during_suppression = 0
        self._enabled = True
        self._in_component_guard = BooleanContext()
        self._accumulate_midi_messages = BooleanContext()
        self._midi_message_dict = {}
        self._midi_message_list = []
        self._midi_message_count = 0
        self.mxd_midi_scheduler = MessageScheduler(self._do_send_midi, self._task_group.add(Task.TimedCallbackTask()))
        self._control_surface_injector = inject(
            parent_task_group=(const(self._task_group)),
            show_message=(const(self.show_message)),
            log_message=(const(self.log_message)),
            register_component=(const(self._register_component)),
            register_control=(const(self._register_control)),
            request_rebuild_midi_map=(const(self.request_rebuild_midi_map)),
            set_pad_translations=(const(self.set_pad_translations)),
            send_midi=(const(self._send_midi)),
            song=(self.song),
        ).everywhere()
        self.register_slot(self.song(), self._on_track_list_changed, "visible_tracks")
        self.register_slot(self.song(), self._on_scene_list_changed, "scenes")
        self.register_slot(self.song().view, self._on_selected_track_changed, "selected_track")
        self.register_slot(self.song().view, self._on_selected_scene_changed, "selected_scene")

    @property
    def components(self):
        return tuple([comp for comp in self._components if not comp.is_private])

    @property
    def root_components(self):
        return tuple([comp for comp in self._components if comp.is_root if not comp.is_private])

    def _get_tasks(self):
        return self._task_group

    _tasks = property(_get_tasks)

    def application(self):
        return Live.Application.get_application()

    def song(self):
        return self._c_instance.song()

    def disconnect(self):
        self.notify_disconnect(self)
        self._disconnect_and_unregister_all_components()
        with self.component_guard():
            for control in self.controls:
                control.disconnect()
                control.canonical_parent = None

        self._forwarding_registry = None
        self.controls = None
        self._displays = None
        self._device_component = None
        self._pad_translations = None
        cs_list = self._control_surfaces()
        if self in cs_list:
            cs_list.remove(self)
        self._task_group.clear()
        super().disconnect()

    def _control_surfaces(self):
        return get_control_surfaces()

    def can_lock_to_devices(self):
        return self._device_component is not None

    @_scheduled_method
    def lock_to_device(self, device):
        with self.component_guard():
            self._device_component.set_lock_to_device(True, device)

    @_scheduled_method
    def unlock_from_device(self, device):
        with self.component_guard():
            self._device_component.set_lock_to_device(False, device)

    @_scheduled_method
    def restore_bank(self, bank_index):
        with self.component_guard():
            self._device_component.restore_bank(bank_index)

    def suggest_input_port(self):
        return self._suggested_input_port

    def suggest_output_port(self):
        return self._suggested_output_port

    def suggest_map_mode(self, cc_no, channel):
        suggested_map_mode = -1
        for control in self.controls:
            if isinstance(control, InputControlElement):
                if control.message_type() == MIDI_CC_TYPE:
                    if control.message_identifier() == cc_no:
                        if control.message_channel() == channel:
                            suggested_map_mode = control.message_map_mode()
                            break

        return suggested_map_mode

    def suggest_needs_takeover(self, cc_no, channel):
        needs_takeover = True
        for control in self._controls:
            if isinstance(control, InputControlElement):
                if control.message_type() == MIDI_CC_TYPE:
                    if control.message_identifier() == cc_no:
                        if control.message_channel() == channel:
                            needs_takeover = control.needs_takeover()
                            break

        return needs_takeover

    def supports_pad_translation(self):
        return self._pad_translations is not None

    def set_highlighting_session_component(self, session_component):
        if self._highlighting_session_component is not None:
            self._set_session_highlight(-1, -1, -1, -1, False)
            self._highlighting_session_component.set_highlighting_callback(None)
        if session_component is not None:
            session_component.set_highlighting_callback(self._set_session_highlight)
        self._highlighting_session_component = session_component

    def highlighting_session_component(self):
        return self._highlighting_session_component

    def show_message(self, message):
        self._c_instance.show_message(message)

    def log_message(self, *message):
        message = "({}) {}".format(self.__class__.__name__, " ".join(map(str, message)))
        console_message = "LOG: " + message
        logger.info(console_message)
        if self._c_instance:
            self._c_instance.log_message(message)

    def instance_identifier(self):
        return self._c_instance.instance_identifier()

    def connect_script_instances(self, instanciated_scripts):
        pass

    def request_rebuild_midi_map(self):
        if self._suppress_requests_counter > 0:
            self._rebuild_requests_during_suppression += 1
        else:
            self._c_instance.request_rebuild_midi_map()

    def build_midi_map(self, midi_map_handle):
        with self._in_build_midi_map():
            self._forwarding_registry.clear()
            self._forwarding_long_identifier_registry.clear()
            for control in self.controls:
                if isinstance(control, InputControlElement):
                    control.install_connections(
                        self._translate_message,
                        partial(self._install_mapping, midi_map_handle),
                        partial(self._install_forwarding, midi_map_handle),
                    )

            if self._pad_translations is not None:
                self._c_instance.set_pad_translation(self._pad_translations)

    def toggle_lock(self):
        self._c_instance.toggle_lock()

    def port_settings_changed(self):
        self.refresh_state()

    def refresh_state(self):
        self.update()

    def update(self):
        with self.component_guard():
            for control in self.controls:
                control.clear_send_cache()

            for component in self._components:
                component.update()

    @profile
    def update_display(self):
        with self.component_guard():
            with self._is_sending_scheduled_messages():
                self._task_group.update(Defaults.TIMER_DELAY)

    @profile
    def receive_midi(self, midi_bytes):
        with self.component_guard():
            self._do_receive_midi(midi_bytes)

    def is_sysex_message(self, midi_bytes):
        return len(midi_bytes) != 3

    def _do_receive_midi(self, midi_bytes):
        (self.notify_received_midi)(*midi_bytes)
        self.mxd_midi_scheduler.handle_message(midi_bytes)
        if not self.is_sysex_message(midi_bytes):
            self.handle_nonsysex(midi_bytes)
        else:
            self.handle_sysex(midi_bytes)

    def get_recipient_for_nonsysex_midi_message(self, midi_bytes):
        is_pitchbend = midi_bytes[0] & 240 == MIDI_PB_STATUS
        forwarding_key = midi_bytes[: 1 if is_pitchbend else 2]
        if forwarding_key in self._forwarding_registry:
            return self._forwarding_registry[forwarding_key]
        return None

    def handle_nonsysex(self, midi_bytes):
        is_pitchbend = midi_bytes[0] & 240 == MIDI_PB_STATUS
        value = midi_bytes[1] + (midi_bytes[2] << 7) if is_pitchbend else midi_bytes[2]
        recipient = self.get_recipient_for_nonsysex_midi_message(midi_bytes)
        if recipient is not None:
            recipient.receive_value(value)
        else:
            if self.received_midi_listener_count() == 0:
                self.log_message("Got unknown message: " + str(midi_bytes))

    def handle_sysex(self, midi_bytes):
        result = find_if(
            lambda id__: midi_bytes[: len(id__[0])] == id__[0],
            iter(self._forwarding_long_identifier_registry.items()),
        )
        if result is not None:
            id, control = result
            control.receive_value(midi_bytes[len(id) : -1])
        else:
            if self.received_midi_listener_count() == 0:
                self.log_message("Got unknown sysex message: " + str(midi_bytes))

    def set_device_component(self, device_component):
        if self._device_component is not None:
            self._device_component.set_lock_callback(None)
        self._device_component = device_component
        self._c_instance.update_locks()
        if device_component is not None:
            device_component.set_lock_callback(self._toggle_lock)
            if device_component.device_selection_follows_track_selection:
                device_component.update_device_selection()

    @contextmanager
    def suppressing_rebuild_requests(self):
        try:
            self._set_suppress_rebuild_requests(True)
            yield
        finally:
            self._set_suppress_rebuild_requests(False)

    def _set_suppress_rebuild_requests(self, suppress_requests):
        if suppress_requests:
            self._suppress_requests_counter += 1
        else:
            self._suppress_requests_counter -= 1
            if self._suppress_requests_counter == 0:
                if self._rebuild_requests_during_suppression > 0:
                    self.request_rebuild_midi_map()
                    self._rebuild_requests_during_suppression = 0

    def set_pad_translations(self, pad_translations):
        def check_translation(translation):
            return True

        self._pad_translations = pad_translations

    def set_enabled(self, enable):
        bool_enable = bool(enable)
        if self._enabled != bool_enable:
            with self.component_guard():
                self._enabled = bool_enable
                root_components = self.root_components
                components = root_components if len(root_components) > 0 else self._components
                for component in components:
                    component._set_enabled_recursive(bool_enable)

    def schedule_message(self, delay_in_ticks, callback, parameter=None):
        if not self._is_sending_scheduled_messages:
            delay_in_ticks -= 1
        message_reference = [None]

        def message(delta):
            if parameter:
                callback(parameter)
            else:
                callback()
            self._remaining_scheduled_messages.remove(message_reference)

        message_reference[0] = message
        self._remaining_scheduled_messages.append(message_reference)
        if delay_in_ticks:
            self._task_group.add(Task.sequence(Task.delay(delay_in_ticks), message))
        else:
            self._task_group.add(message)

    def _process_remaining_scheduled_messages(self):
        current_scheduled_messages = tuple(self._remaining_scheduled_messages)
        for (message,) in current_scheduled_messages:
            message(None)

    def set_feedback_channels(self, channels):
        self._c_instance.set_feedback_channels(channels)

    def set_controlled_track(self, track):
        self._c_instance.set_controlled_track(track)

    def release_controlled_track(self):
        self._c_instance.release_controlled_track()

    def _register_control(self, control):
        self.controls.append(control)
        control.canonical_parent = self
        if isinstance(control, PhysicalDisplayElement):
            self._displays.append(control)

    def _register_component(self, component):
        self._components.append(component)
        component.canonical_parent = self

    def _disconnect_and_unregister_all_components(self):
        with self.component_guard():
            for component in self._components:
                component.canonical_parent = None
                component.disconnect()

        self._components = []
        self.set_highlighting_session_component(None)
        self.set_device_component(None)

    @contextmanager
    def component_guard(self):
        if not self._in_component_guard:
            with self._in_component_guard():
                with self._component_guard():
                    yield
        else:
            yield

    @property
    def in_component_guard(self):
        return bool(self._in_component_guard)

    @contextmanager
    def _component_guard(self):
        with self._control_surface_injector:
            with self.suppressing_rebuild_requests():
                with self.accumulating_midi_messages():
                    yield

    @profile
    def call_listeners(self, listeners):
        with self.component_guard():
            for listener in filter(lambda l: l is not None, listeners):
                listener()

    @contextmanager
    def accumulating_midi_messages(self):
        with self._accumulate_midi_messages():
            try:
                yield
            finally:
                self._flush_midi_messages()

    def get_control_by_name(self, control_name):
        return find_if(lambda c: c.name == control_name, self.controls)

    def _send_midi(self, midi_event_bytes, optimized=True):
        if self._accumulate_midi_messages:
            sysex_status_byte = 240
            entry = (self._midi_message_count, midi_event_bytes)
            if optimized and midi_event_bytes[0] != sysex_status_byte:
                self._midi_message_dict[(midi_event_bytes[0], midi_event_bytes[1])] = entry
            else:
                self._midi_message_list.append(entry)
            self._midi_message_count += 1
        else:
            self._do_send_midi(midi_event_bytes)
        return True

    def _flush_midi_messages(self):
        for _, message in sorted((chain(self._midi_message_list, iter(self._midi_message_dict.values()))), key=first):
            self._do_send_midi(message)

        self._midi_message_dict.clear()
        self._midi_message_list[:] = []
        self._midi_message_count = 0

    def _do_send_midi(self, midi_event_bytes):
        try:
            self._c_instance.send_midi(midi_event_bytes)
        except:
            self.log_message("Error while sending midi message", midi_event_bytes)
            traceback.print_exc()
            return False
        else:
            return True

    def _install_mapping(self, midi_map_handle, control, parameter, feedback_delay, feedback_map):
        success = False
        feedback_rule = None
        if control.message_type() is MIDI_NOTE_TYPE:
            feedback_rule = Live.MidiMap.NoteFeedbackRule()
            feedback_rule.note_no = control.message_identifier()
            feedback_rule.vel_map = feedback_map
        else:
            if control.message_type() is MIDI_CC_TYPE:
                feedback_rule = Live.MidiMap.CCFeedbackRule()
                feedback_rule.cc_no = control.message_identifier()
                feedback_rule.cc_value_map = feedback_map
            else:
                if control.message_type() is MIDI_PB_TYPE:
                    feedback_rule = Live.MidiMap.PitchBendFeedbackRule()
                    feedback_rule.value_pair_map = feedback_map
        feedback_rule.channel = control.message_channel()
        feedback_rule.delay_in_ms = feedback_delay
        if control.message_type() is MIDI_NOTE_TYPE:
            success = Live.MidiMap.map_midi_note_with_feedback_map(
                midi_map_handle,
                parameter,
                control.message_channel(),
                control.message_identifier(),
                feedback_rule,
            )
        else:
            if control.message_type() is MIDI_CC_TYPE:
                success = Live.MidiMap.map_midi_cc_with_feedback_map(
                    midi_map_handle,
                    parameter,
                    control.message_channel(),
                    control.message_identifier(),
                    control.message_map_mode(),
                    feedback_rule,
                    not control.needs_takeover(),
                    control.mapping_sensitivity,
                )
            else:
                if control.message_type() is MIDI_PB_TYPE:
                    success = Live.MidiMap.map_midi_pitchbend_with_feedback_map(
                        midi_map_handle,
                        parameter,
                        control.message_channel(),
                        feedback_rule,
                        not control.needs_takeover(),
                    )
        if success:
            Live.MidiMap.send_feedback_for_parameter(midi_map_handle, parameter)
        return success

    def _install_forwarding(self, midi_map_handle, control):
        success = False
        if control.message_type() is MIDI_NOTE_TYPE:
            success = Live.MidiMap.forward_midi_note(
                self._c_instance.handle(),
                midi_map_handle,
                control.message_channel(),
                control.message_identifier(),
            )
        else:
            if control.message_type() is MIDI_CC_TYPE:
                success = Live.MidiMap.forward_midi_cc(
                    self._c_instance.handle(),
                    midi_map_handle,
                    control.message_channel(),
                    control.message_identifier(),
                )
            else:
                if control.message_type() is MIDI_PB_TYPE:
                    success = Live.MidiMap.forward_midi_pitchbend(
                        self._c_instance.handle(),
                        midi_map_handle,
                        control.message_channel(),
                    )
                else:
                    success = True
        if success:
            forwarding_keys = control.identifier_bytes()
            for key in forwarding_keys:
                registry = (
                    self._forwarding_registry
                    if control.message_type() != MIDI_SYSEX_TYPE
                    else self._forwarding_long_identifier_registry
                )
                registry[key] = control

        return success

    def _translate_message(self, type, from_identifier, from_channel, to_identifier, to_channel):
        if type == MIDI_CC_TYPE:
            self._c_instance.set_cc_translation(from_identifier, from_channel, to_identifier, to_channel)
        else:
            if type == MIDI_NOTE_TYPE:
                self._c_instance.set_note_translation(from_identifier, from_channel, to_identifier, to_channel)
            else:
                pass

    def _set_session_highlight(self, track_offset, scene_offset, width, height, include_return_tracks):
        if [track_offset, scene_offset, width, height].count(-1) != 4:
            pass
        self._c_instance.set_session_highlight(track_offset, scene_offset, width, height, include_return_tracks)

    def _on_track_list_changed(self):
        for component in self._components:
            component.on_track_list_changed()

        self.schedule_message(1, self._on_selected_track_changed)

    def _on_scene_list_changed(self):
        for component in self._components:
            component.on_scene_list_changed()

    def _on_selected_track_changed(self):
        for component in self._components:
            component.on_selected_track_changed()

    def _on_selected_scene_changed(self):
        for component in self._components:
            component.on_selected_scene_changed()

    def _toggle_lock(self):
        self._c_instance.toggle_lock()

    def _refresh_displays(self):
        for display in self._displays:
            display.update()
            display._tasks.update(Defaults.TIMER_DELAY)


class OptimizedControlSurface(ControlSurface):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._optimized_ownership_handler = OptimizedOwnershipHandler()
        injecting = inject(element_ownership_handler=(const(self._optimized_ownership_handler)))
        self._ownership_handler_injector = injecting.everywhere()

    @contextmanager
    def _component_guard(self):
        with super()._component_guard():
            with self._ownership_handler_injector:
                yield
                self._optimized_ownership_handler.commit_ownership_changes()

    def _register_control(self, control):
        super()._register_control(control)
        if old_hasattr(control, "_is_resource_based"):
            control._is_resource_based = True



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/playable.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4887 bytes
from functools import partial

from ...base import listenable_property
from .. import Component
from ..controls import ButtonControl, PlayableControl, control_matrix


class PlayableComponent(Component):
    matrix = control_matrix(PlayableControl, color=None)
    select_button = ButtonControl(color=None)
    pressed_pads = listenable_property.managed([])

    def __init__(self, name="Playable", matrix_always_listenable=False, *a, **k):
        (super().__init__)(a, name=name, **k)
        self._takeover_pads = False
        self._default_playable_mode = (
            PlayableControl.Mode.playable_and_listenable if matrix_always_listenable else PlayableControl.Mode.playable
        )
        self.matrix._control_type = partial(PlayableControl, mode=(self._default_playable_mode))

    @property
    def width(self):
        if self.matrix.width:
            return self.matrix.width
        return 4

    @property
    def height(self):
        if self.matrix.height:
            return self.matrix.height
        return 4

    def set_matrix(self, matrix):
        self.matrix.set_control_element(matrix)
        self._reset_selected_pads()
        self._update_led_feedback()
        self._update_note_translations()

    def _set_control_pads_from_script(self, takeover_pads):
        if takeover_pads != self._takeover_pads:
            self._takeover_pads = takeover_pads
            self._update_control_from_script()

    def _update_control_from_script(self):
        takeover_pads = self._takeover_pads or len(self.pressed_pads) > 0
        mode = PlayableControl.Mode.listenable if takeover_pads else self._default_playable_mode
        for button in self.matrix:
            button.set_mode(mode)

    @matrix.pressed
    def matrix(self, button):
        self._on_matrix_pressed(button)

    @matrix.released
    def matrix(self, button):
        self._on_matrix_released(button)

    def _on_matrix_pressed(self, button):
        self.pressed_pads = [*self.pressed_pads, button]
        if len(self.pressed_pads) == 1:
            self._update_control_from_script()

    def _on_matrix_released(self, button):
        if button in self.pressed_pads:
            self.pressed_pads = [p for p in self.pressed_pads if p is not button]
            if not self.pressed_pads:
                self._update_control_from_script()
        self._update_led_feedback()

    @select_button.value
    def select_button(self, _, button):
        self._set_control_pads_from_script(button.is_pressed)

    def _update_led_feedback(self):
        for button in self.matrix:
            self._update_button_color(button)

    def _update_button_color(self, button):
        pass

    def _update_note_translations(self):
        for button in self.matrix:
            if self._button_should_be_enabled(button):
                self._set_button_control_properties(button)
                button.enabled = True
            else:
                button.enabled = False

    def _reset_selected_pads(self):
        if self.pressed_pads:
            self.pressed_pads = []
            self._update_control_from_script()

    def _set_button_control_properties(self, button):
        identifier, channel = self._note_translation_for_button(button)
        button.identifier = identifier
        button.channel = channel

    def _button_should_be_enabled(self, button):
        identifier, _ = self._note_translation_for_button(button)
        return (identifier is None) or ((isinstance(identifier, int)) and (identifier < 128))

    def _note_translation_for_button(self, button):
        return (button.identifier, button.channel)

    def update(self):
        super().update()
        if self.is_enabled():
            self._set_control_pads_from_script(False)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/compound_element.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 11953 bytes
from collections import OrderedDict

from future.utils import iteritems

from ..base import BooleanContext, first, listens_group, second
from .control_element import ControlElementClient, NotifyingControlElement


class NestedElementClient(ControlElementClient):
    def __init__(self, compound=None, client=None, **k):
        (super().__init__)(**k)
        self.compound = compound
        self.client = client

    def set_control_element(self, element, grabbed):
        self.compound.set_control_element(element, grabbed)


class CompoundElement(NotifyingControlElement, ControlElementClient):
    _is_resource_based = False

    def __init__(self, control_elements=None, *a, **k):
        (super().__init__)(*a, **k)
        resource = self.resource
        original_grab_resource = resource.grab
        original_release_resource = resource.release

        def grab_resource(client, **k):
            (self._grab_nested_control_elements)(client, **k)
            original_grab_resource(client, **k)

        def release_resource(client):
            original_release_resource(client)
            self._release_nested_control_elements(client)

        self.resource.grab = grab_resource
        self.resource.release = release_resource
        self._nested_control_elements = OrderedDict()
        self._nested_element_clients = {}
        self._disable_notify_owner_on_button_ownership_change = BooleanContext()
        self._listen_nested_requests = 0
        if control_elements is not None:
            (self.register_control_elements)(*control_elements)

    def on_nested_control_element_received(self, control):
        pass

    def on_nested_control_element_lost(self, control):
        pass

    def on_nested_control_element_value(self, value, control):
        self.notify_value(value, control)

    def get_control_element_priority(self, element, priority):
        return priority

    def register_control_elements(self, *elements):
        return list(map(self.register_control_element, elements))

    def register_control_element(self, element):
        self._nested_control_elements[element] = False
        if self._listen_nested_requests > 0:
            self._CompoundElement__on_nested_control_element_value.add_subject(element)
        if self._is_resource_based and self.resource.owner:
            priority = self.get_control_element_priority(element, self.resource.max_priority)
            nested_client = self._get_nested_client(self.resource.owner)
            element.resource.grab(nested_client, priority=priority)
        else:
            if not self._is_resource_based:
                with self._disable_notify_owner_on_button_ownership_change():
                    element.notify_ownership_change(self, True)
        return element

    def unregister_control_elements(self, *elements):
        return list(map(self.unregister_control_element, elements))

    def unregister_control_element(self, element):
        if self._is_resource_based:
            for client in self.resource.clients:
                nested_client = self._get_nested_client(client)
                element.resource.release(nested_client)

        else:
            with self._disable_notify_owner_on_button_ownership_change():
                element.notify_ownership_change(self, False)
        if self._listen_nested_requests > 0:
            self._CompoundElement__on_nested_control_element_value.remove_subject(element)
        del self._nested_control_elements[element]
        return element

    def has_control_element(self, control):
        return control in self._nested_control_elements

    def owns_control_element(self, control):
        return self._nested_control_elements.get(control, False)

    def owned_control_elements(self):
        return list(map(first, filter(second, iteritems(self._nested_control_elements))))

    def nested_control_elements(self):
        return list(self._nested_control_elements)

    def reset(self):
        for element in self.owned_control_elements():
            element.reset()

    def reset_state(self):
        for element in self.owned_control_elements():
            element.reset_state()

    def add_value_listener(self, *a, **k):
        if self.value_listener_count() == 0:
            self.request_listen_nested_control_elements()
        (super().add_value_listener)(*a, **k)

    def remove_value_listener(self, *a, **k):
        (super().remove_value_listener)(*a, **k)
        if self.value_listener_count() == 0:
            self.unrequest_listen_nested_control_elements()

    def request_listen_nested_control_elements(self):
        if self._listen_nested_requests == 0:
            self._connect_nested_control_elements()
        self._listen_nested_requests += 1

    def unrequest_listen_nested_control_elements(self):
        if self._listen_nested_requests == 1:
            self._disconnect_nested_control_elements()
        self._listen_nested_requests -= 1

    def _connect_nested_control_elements(self):
        self._CompoundElement__on_nested_control_element_value.replace_subjects(
            list(self._nested_control_elements.keys()),
        )

    def _disconnect_nested_control_elements(self):
        self._CompoundElement__on_nested_control_element_value.replace_subjects([])

    def _on_nested_control_element_received(self, control):
        if control in self._nested_control_elements:
            if not self._nested_control_elements[control]:
                self._nested_control_elements[control] = True
        self.on_nested_control_element_received(control)

    def _on_nested_control_element_lost(self, control):
        if control in self._nested_control_elements:
            if self._nested_control_elements[control]:
                self._nested_control_elements[control] = False
        self.on_nested_control_element_lost(control)

    @listens_group("value")
    def __on_nested_control_element_value(self, value, sender):
        if self.owns_control_element(sender):
            self.on_nested_control_element_value(value, sender)

    def set_control_element(self, control, grabbed):
        if grabbed:
            self._on_nested_control_element_received(control)
        else:
            self._on_nested_control_element_lost(control)
        owner = self._resource.owner
        if owner:
            if not self._disable_notify_owner_on_button_ownership_change:
                self.notify_ownership_change(owner, True)

    def _grab_nested_control_elements(self, client, priority=None, **k):
        was_resource_based = self._is_resource_based
        self._is_resource_based = True
        nested_client = self._get_nested_client(client)
        with self._disable_notify_owner_on_button_ownership_change():
            for element in list(self._nested_control_elements.keys()):
                if not was_resource_based:
                    element.notify_ownership_change(self, False)
                else:
                    nested_priority = self.get_control_element_priority(element, priority)
                    element.resource.grab(nested_client, priority=nested_priority)

    def _release_nested_control_elements(self, client):
        nested_client = self._get_nested_client(client)
        with self._disable_notify_owner_on_button_ownership_change():
            for element in list(self._nested_control_elements.keys()):
                element.resource.release(nested_client)

    def _get_nested_client(self, client):
        try:
            nested_client = self._nested_element_clients[client]
        except KeyError:
            nested_client = self._nested_element_clients[client] = NestedElementClient(self, client)

        return nested_client

    def __len__(self):
        return len(self._nested_control_elements)

    def __iter__(self):
        for element, owned in iteritems(self._nested_control_elements):
            yield element if owned else None

    def __getitem__(self, index_or_slice):
        if isinstance(index_or_slice, slice):
            items = list(self._nested_control_elements.items())[index_or_slice]
            return [element if owned else None for element, owned in items]
        element, owned = list(self._nested_control_elements.items())[index_or_slice]
        if owned:
            return element
        return None



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/target_track.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3202 bytes
from ableton.v2.base import listens, listens_group, liveobj_valid
from ableton.v2.control_surface import Component


class TargetTrackComponent(Component):
    __events__ = ("target_track",)

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._target_track = None
        self._armed_track_list = []
        self._TargetTrackComponent__on_selected_track_changed.subject = self.song.view
        self._TargetTrackComponent__on_selected_track_changed()

    @property
    def target_track(self):
        return self._target_track

    @listens("selected_track")
    def __on_selected_track_changed(self):
        if not self._armed_track_list:
            self._set_target_track()

    def _set_target_track(self):
        new_target = self._target_track
        new_target = self._armed_track_list[-1] if self._armed_track_list else self.song.view.selected_track
        if self._target_track != new_target:
            self._target_track = new_target
            self.notify_target_track()


class ArmedTargetTrackComponent(TargetTrackComponent):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._ArmedTargetTrackComponent__on_tracks_changed.subject = self.song
        self._ArmedTargetTrackComponent__on_tracks_changed()

    @property
    def tracks(self):
        return list(filter(lambda t: liveobj_valid(t) and t.can_be_armed and t.has_midi_input, self.song.tracks))

    @listens("visible_tracks")
    def __on_tracks_changed(self):
        tracks = self.tracks
        self._ArmedTargetTrackComponent__on_arm_changed.replace_subjects(tracks)
        self._ArmedTargetTrackComponent__on_frozen_state_changed.replace_subjects(tracks)
        self._refresh_armed_track_list()

    @listens_group("arm")
    def __on_arm_changed(self, _):
        self._refresh_armed_track_list()

    @listens_group("is_frozen")
    def __on_frozen_state_changed(self, _):
        self._refresh_armed_track_list()

    def _refresh_armed_track_list(self):
        tracks = self.tracks
        for track in self._armed_track_list:
            if liveobj_valid(track):
                if track.arm:
                    if not track.is_frozen:
                        if track not in tracks:
                            pass
                    self._armed_track_list.remove(track)

        for track in tracks:
            if track.arm:
                if not track.is_frozen:
                    if track not in self._armed_track_list:
                        self._armed_track_list.append(track)

        self._set_target_track()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/ButtonMatrixElement.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4231 bytes
from .CompoundElement import CompoundElement
from .Util import const, product, slicer, to_slice


class ButtonMatrixElement(CompoundElement):
    def __init__(self, rows=None, *a, **k):
        if rows is None:
            rows = []
        (super().__init__)(*a, **k)
        self._buttons = []
        self._orig_buttons = []
        self._button_coordinates = {}
        self._max_row_width = 0
        list(map(self.add_row, rows))

    @property
    @slicer(2)
    def submatrix(self, col_slice, row_slice):
        col_slice = to_slice(col_slice)
        row_slice = to_slice(row_slice)
        rows = [row[col_slice] for row in self._orig_buttons[row_slice]]
        return ButtonMatrixElement(rows=rows)

    def add_row(self, buttons):
        self._buttons.append([None] * len(buttons))
        self._orig_buttons.append(buttons)
        for index, button in enumerate(buttons):
            self._button_coordinates[button] = (index, len(self._buttons) - 1)
            self.register_control_element(button)

        if self._max_row_width < len(buttons):
            self._max_row_width = len(buttons)

    def width(self):
        return self._max_row_width

    def height(self):
        return len(self._buttons)

    def send_value(self, column, row, value, force=False):
        if len(self._buttons[row]) > column:
            button = self._buttons[row][column]
            if button:
                button.send_value(value, force=force)

    def set_light(self, column, row, value):
        if len(self._buttons[row]) > column:
            button = self._buttons[row][column]
            if button:
                button.set_light(value)

    def get_button(self, column, row):
        if len(self._buttons[row]) > column:
            return self._buttons[row][column]
        return None

    def reset(self):
        for button in self:
            if button:
                button.reset()

    def __iter__(self):
        for j, i in product(range(self.height()), range(self.width())):
            button = self.get_button(i, j)
            yield button

    def __getitem__(self, index):
        if isinstance(index, slice):
            indices = index.indices(len(self))
            return list(map(self._do_get_item, list(range(*indices))))
        if index < 0:
            index += len(self)
        return self._do_get_item(index)

    def _do_get_item(self, index):
        row, col = divmod(index, self.width())
        return self.get_button(col, row)

    def __len__(self):
        return self.width() * self.height()

    def iterbuttons(self):
        for j, i in product(range(self.height()), range(self.width())):
            button = self.get_button(i, j)
            yield (button, (i, j))

    def on_nested_control_element_value(self, value, sender):
        x, y = self._button_coordinates[sender]
        is_momentary = getattr(sender, "is_momentary", const(None))()
        self.notify_value(value, x, y, is_momentary)

    def on_nested_control_element_received(self, control):
        x, y = self._button_coordinates[control]
        self._buttons[y][x] = control

    def on_nested_control_element_lost(self, control):
        x, y = self._button_coordinates[control]
        self._buttons[y][x] = None



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/display_specification.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 8114 bytes
import sys
from contextlib import contextmanager
from io import StringIO
from typing import Callable, Optional, Type

from ...base import Disconnectable, const, inject
from .notifications import DefaultNotifications, Notifications
from .renderable import Renderable
from .state import State
from .type_decl import DISCONNECT_EVENT, INIT_EVENT
from .util import updating_display
from .view import View


class DisplaySpecification:
    def __init__(
        self,
        create_root_view: Optional[Callable[([], View)]] = None,
        protocol=None,
        notifications: Optional[Type[Notifications]] = None,
    ):
        self.create_root_view = create_root_view
        self.protocol = protocol
        self.notifications = notifications or DefaultNotifications


class Display(Disconnectable):
    def __init__(self, specification: DisplaySpecification, renderable_components, elements):
        self._react_fns = []
        with inject(register_react_fn=(const(self._react_fns.append))).everywhere():
            self._root_view = specification.create_root_view()
        self._display_protocol = specification.protocol
        self._display_fn = None
        self._initialized = False
        self._last_displayed_content = None
        self._is_deferring_render_and_update = False
        self._was_updated_in_deferred_context = False
        self._exception_raised = False
        self._suppress_stdout_from_render = True
        self._state = State()
        for component in renderable_components:
            if component.include_in_top_level_state:
                setattr(self._state, component.name.lower(), component.renderable_state)

        self._state.elements = State()
        for name, element in vars(elements).items():
            if isinstance(element, Renderable):
                setattr(self._state.elements, name, element.renderable_state)

        self._display_fn = self._display_protocol(elements)
        self._initialized = True
        self.react(INIT_EVENT)

    @property
    def state(self):
        return self._state

    def react(self, event):
        if self._initialized:
            for fn in self._react_fns:
                fn(self._state, event)

            self._suppress_stdout_from_render = False
            self.render_and_update_display()

    def render(self):
        return self._root_view(self._state)

    def display(self, content):
        with updating_display():
            self._display_fn(content)

    def render_and_update_display(self):
        try:
            if self._is_deferring_render_and_update:
                self._was_updated_in_deferred_context = True
            else:
                if self._display_fn:
                    captured_stdout = StringIO()
                    with capturing_stdout(captured_stdout):
                        content = self.render()
                    if content != self._last_displayed_content:
                        self.display(content)
                        self._last_displayed_content = content
                        self._suppress_stdout_from_render = False
                    if not self._suppress_stdout_from_render:
                        last_output = captured_stdout.getvalue()
                        if last_output:
                            print(last_output)
                        self._suppress_stdout_from_render = True
        except Exception as e:
            try:
                exception_already_raised = self._exception_raised
                self._exception_raised = True
                if not exception_already_raised:
                    raise e
            finally:
                e = None
                del e

    @contextmanager
    def deferring_render_and_update_display(self):
        self._is_deferring_render_and_update = True
        self._was_updated_in_deferred_context = False
        try:
            yield
        finally:
            self._is_deferring_render_and_update = False
            if self._was_updated_in_deferred_context:
                self.render_and_update_display()

    @property
    def rendered_content(self):
        return self._last_displayed_content

    def disconnect(self):
        self.react(DISCONNECT_EVENT)

    def clear_content_cache(self):
        self._last_displayed_content = None


@contextmanager
def capturing_stdout(temp_output_stream):
    original_stdout = sys.stdout
    sys.stdout = temp_output_stream
    try:
        yield
    finally:
        sys.stdout = original_stdout



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/elements/button.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4323 bytes
import Live
from past.builtins import long

from ...base import BooleanContext, const, has_event, in_range, listens, old_hasattr
from ..input_control_element import InputControlElement
from ..skin import Skin
from .color import Color


class DummyUndoStepHandler:
    def begin_undo_step(self):
        pass

    def end_undo_step(self):
        pass


class ButtonElementMixin:
    def set_light(self, value):
        pass


class ButtonElement(InputControlElement, ButtonElementMixin):
    class ProxiedInterface(InputControlElement.ProxiedInterface, ButtonElementMixin):
        is_momentary = const(True)
        is_pressed = const(False)

    class Colors:
        class DefaultButton:
            On = Color(127)
            Off = Color(1)
            Disabled = Color(0)

    num_delayed_messages = 2

    def __init__(
        self,
        is_momentary,
        msg_type,
        channel,
        identifier,
        is_rgb=False,
        skin=Skin(Colors),
        undo_step_handler=DummyUndoStepHandler(),
        send_should_depend_on_forwarding=False,
        *a,
        **k,
    ):
        (super().__init__)(
            msg_type,
            channel,
            identifier,
            *a,
            send_should_depend_on_forwarding=send_should_depend_on_forwarding,
            **k,
        )
        self.is_rgb = is_rgb
        self._is_momentary = bool(is_momentary)
        self._last_received_value = -1
        self._undo_step_handler = undo_step_handler
        self._skin = skin
        self._drawing_via_skin = BooleanContext()

    def reset(self):
        self.set_light("DefaultButton.Disabled")
        self.use_default_message()
        self.suppress_script_forwarding = False

    def is_momentary(self):
        return self._is_momentary

    def message_map_mode(self):
        return Live.MidiMap.MapMode.absolute

    def is_pressed(self):
        return self._is_momentary and int(self._last_received_value) > 0

    def set_light(self, value):
        if old_hasattr(value, "draw"):
            value.draw(self)
        else:
            if type(value) in (int, long) and in_range(value, 0, 128):
                self.send_value(value)
            else:
                if isinstance(value, bool):
                    self._set_skin_light("DefaultButton.On" if value else "DefaultButton.Off")
                else:
                    self._set_skin_light(value)

    def _set_skin_light(self, value):
        color = None
        try:
            color = self._skin[value]
            self._do_draw(color)
        finally:
            if has_event(color, "midi_value"):
                self._ButtonElement__on_midi_value_changed.subject = color
            else:
                self._disconnect_color_listener()

    def _do_draw(self, color):
        with self._drawing_via_skin():
            color.draw(self)

    @listens("midi_value")
    def __on_midi_value_changed(self, *a):
        self._do_draw(self._ButtonElement__on_midi_value_changed.subject)

    def send_value(self, value, force=False, channel=None):
        if not self._drawing_via_skin:
            self._disconnect_color_listener()
        super().send_value(value, force, channel)

    def receive_value(self, value):
        pressed_before = self.is_pressed()
        self._last_received_value = value
        if not pressed_before:
            if self.is_pressed():
                self._undo_step_handler.begin_undo_step()
        super().receive_value(value)
        if pressed_before:
            if not self.is_pressed():
                self._undo_step_handler.end_undo_step()

    def disconnect(self):
        super().disconnect()
        self._undo_step_handler = None

    def _disconnect_color_listener(self):
        self._ButtonElement__on_midi_value_changed.subject = None



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/event_signal/core.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 524 bytes
from typing import Any

from notifications.type_decl import NOTIFICATION_EVENT_ID

from ..state import State
from ..type_decl import Event
from .type_decl import EventSignalFn


def on_notification() -> EventSignalFn[Any]:
    def signal_fn(_: State, event: Event):
        if event.name == NOTIFICATION_EVENT_ID:
            return event.value
        return None

    return signal_fn



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/base/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1595 bytes
from ableton.v2.base import (
    BooleanContext,
    CompoundDisconnectable,
    Disconnectable,
    EventObject,
    MultiSlot,
    ObservablePropertyAlias,
    SlotGroup,
    chunks,
    clamp,
    compose,
    const,
    depends,
    find_if,
    first,
    flatten,
    forward_property,
    group,
    in_range,
    index_if,
    inject,
    is_iterable,
    lazy_attribute,
    listenable_property,
    listens,
    listens_group,
    memoize,
    mixin,
    nop,
    old_hasattr,
    product,
    recursive_map,
    sign,
    task,
)
from ableton.v2.base.event import EventObjectMeta

from .util import PITCH_NAMES, as_ascii, get_default_ascii_translations, hex_to_rgb, pitch_index_to_string

__all__ = (
    "PITCH_NAMES",
    "BooleanContext",
    "CompoundDisconnectable",
    "Disconnectable",
    "EventObject",
    "EventObjectMeta",
    "MultiSlot",
    "ObservablePropertyAlias",
    "SlotGroup",
    "as_ascii",
    "chunks",
    "clamp",
    "compose",
    "const",
    "depends",
    "find_if",
    "first",
    "flatten",
    "forward_property",
    "get_default_ascii_translations",
    "group",
    "hex_to_rgb",
    "in_range",
    "index_if",
    "inject",
    "is_iterable",
    "lazy_attribute",
    "listenable_property",
    "listens",
    "listens_group",
    "memoize",
    "mixin",
    "nop",
    "old_hasattr",
    "pitch_index_to_string",
    "product",
    "recursive_map",
    "sign",
    "task",
)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/parameter_slot_description.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 6428 bytes
from ableton.v2.base import const

from ..base import EventObject, find_if, listenable_property, listens_group, liveobj_valid

RESULTING_NAME_KEY = "ResultingName"
DISPLAY_NAME_TRANSFORMER_KEY = "DisplayNameTransformer"
CONDITION_NAME_KEY = "ConditionName"
CONDITIONS_LIST_NAME_KEY = "ConditionsListName"
PREDICATE_KEY = "Predicate"
OPERAND_NAME_KEY = "Operand"
AND = "and"
OR = "or"


def find_parameter(name, host):
    parameters = host.parameters if liveobj_valid(host) else []
    return find_if(lambda p: p.original_name == name, parameters)


class ParameterSlotDescription(EventObject):
    __events__ = ("content",)

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._parameter_host = None
        self._default_parameter_name = ""
        self._display_name_transformer = None
        self._conditions = []
        self._cached_content = (None, None)

    def _calc_content(self):
        parameter_name = self._default_parameter_name
        display_name_transformer = self._display_name_transformer
        for condition in self._conditions:
            result = True
            for subcond in condition[CONDITIONS_LIST_NAME_KEY]:
                result = eval(
                    f"{result} {subcond[OPERAND_NAME_KEY]} {subcond[PREDICATE_KEY](find_parameter(subcond[CONDITION_NAME_KEY], self._parameter_host))}",
                )
                if not result:
                    continue

            if result:
                parameter_name = condition[RESULTING_NAME_KEY]
                display_name_transformer = condition[DISPLAY_NAME_TRANSFORMER_KEY]
                break

        return (parameter_name, display_name_transformer)

    @listens_group("value")
    def __on_condition_value_changed(self, _parameter):
        new_content = self._calc_content()
        if new_content != self._cached_content:
            self._cached_content = new_content
            self.notify_content()

    def set_parameter_host(self, host):
        self._parameter_host = host
        self._cached_content = self._calc_content()
        params_names = set()
        for c in self._conditions:
            params_names.update([cond[CONDITION_NAME_KEY] for cond in c[CONDITIONS_LIST_NAME_KEY]])

        self._ParameterSlotDescription__on_condition_value_changed.replace_subjects(
            [find_parameter(name, self._parameter_host) for name in params_names],
        )

    def if_parameter(self, parameter_name):
        self._conditions.append(
            {
                RESULTING_NAME_KEY: self._default_parameter_name,
                DISPLAY_NAME_TRANSFORMER_KEY: self._display_name_transformer,
                CONDITIONS_LIST_NAME_KEY: [{CONDITION_NAME_KEY: parameter_name, OPERAND_NAME_KEY: AND}],
            },
        )
        self._default_parameter_name = ""
        self._display_name_transformer = None
        return self

    def chain_condition(self, operand, parameter_name):
        self._conditions[-1][CONDITIONS_LIST_NAME_KEY].append(
            {CONDITION_NAME_KEY: parameter_name, OPERAND_NAME_KEY: operand},
        )
        return self

    def and_parameter(self, parameter_name):
        return self.chain_condition(AND, parameter_name)

    def or_parameter(self, parameter_name):
        return self.chain_condition(OR, parameter_name)

    def _add_condition_predicate(self, predicate):
        self._conditions[-1][CONDITIONS_LIST_NAME_KEY][-1][PREDICATE_KEY] = predicate

    def has_value(self, value):
        self._add_condition_predicate(
            lambda p: str(p) == value,
        )
        return self

    def has_value_in(self, values):
        self._add_condition_predicate(
            lambda p: str(p) in values,
        )
        return self

    def does_not_have_value(self, value):
        self._add_condition_predicate(
            lambda p: str(p) != value,
        )
        return self

    def is_available(self, value):
        self._add_condition_predicate(
            lambda p: liveobj_valid(p) == value,
        )
        return self

    def else_use(self, parameter_name):
        self._default_parameter_name = parameter_name
        self._display_name_transformer = None
        return self

    def with_name(self, display_name):
        self._display_name_transformer = const(display_name)
        return self

    def with_name_transformer(self, func):
        self._display_name_transformer = func
        return self

    @listenable_property
    def display_name_transformer(self):
        return self._cached_content[1]

    def __str__(self):
        return self._cached_content[0]


def use(parameter_name):
    return ParameterSlotDescription().else_use(parameter_name)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/undo_redo.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1331 bytes
from typing import cast

from .. import Component
from ..controls import ButtonControl
from ..display import Renderable


class UndoRedoComponent(Component, Renderable):
    undo_button = ButtonControl(color="UndoRedo.Undo", pressed_color="UndoRedo.UndoPressed")
    redo_button = ButtonControl(color="UndoRedo.Redo", pressed_color="UndoRedo.RedoPressed")

    def __init__(self, name="Undo_Redo", *a, **k):
        (super().__init__)(a, name=name, **k)

    @undo_button.pressed
    def undo_button(self, _):
        if self.song.can_undo:
            self.notify(self.notifications.UndoRedo.undo, cast(str, self.song.undo()))
        else:
            self.notify(self.notifications.UndoRedo.error_undo_no_action)

    @redo_button.pressed
    def redo_button(self, _):
        if self.song.can_redo:
            self.notify(self.notifications.UndoRedo.redo, cast(str, self.song.redo()))
        else:
            self.notify(self.notifications.UndoRedo.error_redo_no_action)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/capabilities.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 600 bytes
from ableton.v2.control_surface.capabilities import (
    CONTROLLER_ID_KEY,
    HIDDEN,
    NOTES_CC,
    PORTS_KEY,
    REMOTE,
    SCRIPT,
    SYNC,
    controller_id,
    inport,
    outport,
)

__all__ = (
    "CONTROLLER_ID_KEY",
    "HIDDEN",
    "NOTES_CC",
    "PORTS_KEY",
    "REMOTE",
    "SCRIPT",
    "SYNC",
    "controller_id",
    "inport",
    "outport",
)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/device_provider.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 5974 bytes
import Live

from ..base import EventObject, listenable_property, listens, liveobj_changed, liveobj_valid


def device_to_appoint(device):
    appointed_device = device
    if not liveobj_valid(device) or device.can_have_drum_pads:
        if not device.has_macro_mappings:
            if len(device.chains) > 0:
                if liveobj_valid(device.view.selected_chain):
                    if len(device.view.selected_chain.devices) > 0:
                        appointed_device = device_to_appoint(device.view.selected_chain.devices[0])
        return appointed_device
    return None


def select_and_appoint_device(song, device_to_select, ignore_unmapped_macros=True):
    appointed_device = device_to_select
    if ignore_unmapped_macros:
        appointed_device = device_to_appoint(device_to_select)
    song.view.select_device(device_to_select, False)
    song.appointed_device = appointed_device


class DeviceProvider(EventObject):
    device_selection_follows_track_selection = True

    def __init__(self, song=None, *a, **k):
        (super().__init__)(*a, **k)
        self._device = None
        self._locked_to_device = False
        self.song = song
        self._DeviceProvider__on_appointed_device_changed.subject = song
        self._DeviceProvider__on_selected_track_changed.subject = song.view
        self._DeviceProvider__on_selected_device_changed.subject = song.view.selected_track.view

    @listenable_property
    def device(self):
        return self._device

    @device.setter
    def device(self, device):
        if liveobj_changed(self._device, device):
            if not self.is_locked_to_device:
                self._device = device
                self.notify_device()

    @listenable_property
    def is_locked_to_device(self):
        return self._locked_to_device

    def lock_to_device(self, device):
        self.device = device
        self._locked_to_device = True
        self.notify_is_locked_to_device()

    def unlock_from_device(self):
        self._locked_to_device = False
        self.notify_is_locked_to_device()
        self.update_device_selection()

    def update_device_selection(self):
        view = self.song.view
        track_or_chain = view.selected_chain if view.selected_chain else view.selected_track
        device_to_select = None
        if isinstance(track_or_chain, Live.Track.Track):
            device_to_select = track_or_chain.view.selected_device
        if not liveobj_valid(device_to_select):
            if len(track_or_chain.devices) > 0:
                device_to_select = track_or_chain.devices[0]
        if liveobj_valid(device_to_select):
            appointed_device = device_to_appoint(device_to_select)
            self.song.view.select_device(device_to_select, False)
            self.song.appointed_device = appointed_device
            self.device = appointed_device
        else:
            self.song.appointed_device = None
            self.device = None

    def _appoint_device_from_song(self):
        self.device = device_to_appoint(self.song.appointed_device)

    @listens("appointed_device")
    def __on_appointed_device_changed(self):
        self._appoint_device_from_song()

    @listens("has_macro_mappings")
    def __on_has_macro_mappings_changed(self):
        self.song.appointed_device = device_to_appoint(self.song.view.selected_track.view.selected_device)

    @listens("selected_track")
    def __on_selected_track_changed(self):
        self._DeviceProvider__on_selected_device_changed.subject = self.song.view.selected_track.view
        if self.device_selection_follows_track_selection:
            self.update_device_selection()

    @listens("selected_device")
    def __on_selected_device_changed(self):
        self._update_appointed_device()

    @listens("chains")
    def __on_chains_changed(self):
        self._update_appointed_device()

    def _update_appointed_device(self):
        song = self.song
        device = song.view.selected_track.view.selected_device
        if liveobj_valid(device):
            self.song.appointed_device = device_to_appoint(device)
        rack_device = device if isinstance(device, Live.RackDevice.RackDevice) else None
        self._DeviceProvider__on_has_macro_mappings_changed.subject = rack_device
        self._DeviceProvider__on_chains_changed.subject = rack_device



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/clipboard.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 3070 bytes
from ableton.v3.base import listenable_property
from ableton.v3.control_surface import Component
from ableton.v3.control_surface.controls import ButtonControl
from ableton.v3.control_surface.display import Renderable


class ClipboardComponent(Component, Renderable):
    copy_button = ButtonControl(color="Clipboard.Empty", on_color="Clipboard.Filled")
    has_content = listenable_property.managed(False)

    def __init__(self, name="Clipboard", *a, **k):
        (super().__init__)(a, name=name, **k)
        self._source_obj = None
        self._did_paste = False
        self._pending_clear = False

    def set_copy_button(self, button):
        self.clear()
        self.copy_button.set_control_element(button)

    def copy_or_paste(self, obj):
        if self.has_content:
            if self._is_source_valid():
                self._did_paste = self._do_paste(obj)
                if self._did_paste:
                    self.copy_button.is_pressed or self.clear()
            else:
                self.clear(notify=True)
        else:
            self._source_obj = self._do_copy(obj)
            self.update()

    def clear(self, notify=False):
        self._source_obj = None
        self._pending_clear = False
        self.update()
        if notify:
            self.notify(self.notifications.Clipboard.clear)

    @copy_button.pressed
    def copy_button(self, _):
        self._pending_clear = self.has_content

    @copy_button.released
    def copy_button(self, _):
        if self._did_paste or self._pending_clear:
            self.clear(notify=True)

    def update(self):
        super().update()
        self._did_paste = False
        self.has_content = self._source_obj is not None
        self.copy_button.is_on = self.has_content

    def _do_copy(self, obj):
        self._source_obj = obj
        return self._source_obj

    def _do_paste(self, obj):
        self._did_paste = obj is not None
        return self._did_paste

    def _is_source_valid(self):
        return self._source_obj is not None



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/session_recording.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 12909 bytes
import Live

from ...base import EventObject, find_if, listens, liveobj_valid
from ..component import Component
from ..control import ButtonControl, ToggleButtonControl


def track_fired_slot(track):
    index = track.fired_slot_index
    if index >= 0:
        return track.clip_slots[index]
    return None


def track_playing_slot(track):
    index = track.playing_slot_index
    if index >= 0:
        return track.clip_slots[index]
    return None


def track_is_recording(track):
    playing_slot = track_playing_slot(track)
    return playing_slot and playing_slot.is_recording


def track_will_record(track):
    fired_slot = track_fired_slot(track)
    return fired_slot and fired_slot.will_record_on_start


class SessionRecordingComponent(Component):
    automation_button = ToggleButtonControl(toggled_color="Automation.On", untoggled_color="Automation.Off")
    re_enable_automation_button = ButtonControl(color="Automation.On")
    delete_automation_button = ButtonControl(color="Automation.Off")
    record_button = ButtonControl(color="Recording.Off")

    def __init__(self, view_controller=None, *a, **k):
        (super().__init__)(*a, **k)
        self._target_slots = []
        self._view_controller = view_controller
        self._new_button = None
        self._scene_list_new_button = None
        self._length_press_state = None
        self._new_scene_button = None
        song = self.song
        self._SessionRecordingComponent__on_tracks_changed_in_live.subject = song
        self._SessionRecordingComponent__on_is_playing_changed_in_live.subject = song
        self._track_subject_slots = self.register_disconnectable(EventObject())
        self._reconnect_track_listeners()
        self.register_slot(song, self.update, "overdub")
        self.register_slot(song, self.update, "session_record_status")
        self.register_slot(song, self._update_record_button, "session_record")
        self.register_slot(song.view, self.update, "selected_track")
        self.register_slot(song.view, self.update, "selected_scene")
        self.register_slot(song.view, self.update, "detail_clip")
        self._SessionRecordingComponent__on_session_automation_record_changed.subject = song
        self._SessionRecordingComponent__on_session_automation_record_changed()
        self._SessionRecordingComponent__on_re_enable_automation_enabled_changed.subject = song
        self._SessionRecordingComponent__on_re_enable_automation_enabled_changed()

    def set_scene_list_new_button(self, button):
        self._scene_list_new_button = button
        self._SessionRecordingComponent__on_scene_list_new_button_value.subject = button
        self._update_scene_list_new_button()

    def set_new_button(self, button):
        self._new_button = button
        self._SessionRecordingComponent__on_new_button_value.subject = button
        self._update_new_button()

    def set_new_scene_button(self, button):
        self._new_scene_button = button
        self._SessionRecordingComponent__on_new_scene_button_value.subject = button
        self._update_new_scene_button()

    def deactivate_recording(self):
        self._stop_recording()

    @listens("re_enable_automation_enabled")
    def __on_re_enable_automation_enabled_changed(self):
        self.re_enable_automation_button.enabled = self.song.re_enable_automation_enabled

    @re_enable_automation_button.released
    def re_enable_automation_button(self, button):
        if self.song.re_enable_automation_enabled:
            self.song.re_enable_automation()

    @listens("session_automation_record")
    def __on_session_automation_record_changed(self):
        self._on_session_automation_record_changed()

    def _on_session_automation_record_changed(self):
        self.automation_button.is_toggled = self.song.session_automation_record

    @automation_button.toggled
    def automation_button(self, is_toggled, button):
        self.song.session_automation_record = is_toggled

    @delete_automation_button.released
    def delete_automation_button(self, button):
        self._delete_automation_value()

    def update(self):
        super().update()
        if self.is_enabled():
            self._on_playing_clip_has_envelopes_changed.subject = self._get_playing_clip()
            self._update_record_button()
            self._update_new_button()
            self._update_scene_list_new_button()
            self._update_new_scene_button()

    def _update_scene_list_new_button(self):
        self._update_generic_new_button(self._scene_list_new_button)

    def _update_new_button(self):
        self._update_generic_new_button(self._new_button)

    def _update_generic_new_button(self, new_button):
        if new_button:
            if self.is_enabled():
                song = self.song
                selected_track = song.view.selected_track
                clip_slot = song.view.highlighted_clip_slot
                can_new = (
                    (liveobj_valid(clip_slot))
                    and (clip_slot.clip)
                    or (selected_track.can_be_armed and selected_track.playing_slot_index >= 0)
                )
                new_button.set_light("DefaultButton.On" if can_new else "DefaultButton.Disabled")

    def _update_new_scene_button(self):
        if self._new_scene_button:
            if self.is_enabled():
                song = self.song
                track_is_playing = find_if(lambda x: x.playing_slot_index >= 0, song.tracks)
                can_new = not song.view.selected_scene.is_empty or track_is_playing
                self._new_scene_button.set_light("DefaultButton.On" if can_new else "DefaultButton.Disabled")

    def _update_record_button(self):
        if self.is_enabled():
            song = self.song
            status = song.session_record_status
            if status == Live.Song.SessionRecordStatus.transition:
                self.record_button.color = "Recording.Transition"
            else:
                if status == Live.Song.SessionRecordStatus.on or song.session_record:
                    self.record_button.color = "Recording.On"
                else:
                    self.record_button.color = "Recording.Off"

    @listens("has_envelopes")
    def _on_playing_clip_has_envelopes_changed(self):
        self._update_delete_automation_button_color()

    def _update_delete_automation_button_color(self):
        if self._on_playing_clip_has_envelopes_changed.subject:
            self.delete_automation_button.color = (
                "Automation.On"
                if self._on_playing_clip_has_envelopes_changed.subject.has_envelopes
                else "Automation.Off"
            )

    def _delete_automation_value(self):
        if self.is_enabled():
            clip = self._get_playing_clip()
            selected_track = self.song.view.selected_track
            track_frozen = selected_track and selected_track.is_frozen
            if clip:
                if not track_frozen:
                    clip.clear_all_envelopes()
                self._update_delete_automation_button_color()

    def _get_playing_clip(self):
        playing_clip = None
        selected_track = self.song.view.selected_track
        try:
            playing_slot_index = selected_track.playing_slot_index
            if playing_slot_index >= 0:
                playing_clip = selected_track.clip_slots[playing_slot_index].clip
        except (RuntimeError, AttributeError):
            pass

        return playing_clip

    def _handle_limitation_error_on_scene_creation(self):
        pass

    @listens("tracks")
    def __on_tracks_changed_in_live(self):
        self._reconnect_track_listeners()

    @listens("is_playing")
    def __on_is_playing_changed_in_live(self):
        if self.is_enabled():
            self._update_record_button()

    @record_button.pressed
    def record_button(self, button):
        self._on_record_button_pressed()

    def _on_record_button_pressed(self):
        self._trigger_recording()

    def _trigger_recording(self):
        if self.is_enabled():
            if not self._stop_recording():
                self._start_recording()

    @record_button.released
    def record_button(self, button):
        self._on_record_button_released()

    def _on_record_button_released(self):
        pass

    @listens("value")
    def __on_new_scene_button_value(self, value):
        if not self.is_enabled() and value or self._prepare_new_action():
            song = self.song
            selected_scene_index = list(song.scenes).index(song.view.selected_scene)
            try:
                self._create_silent_scene(selected_scene_index)
            except Live.Base.LimitationError:
                self._handle_limitation_error_on_scene_creation()

    @listens("value")
    def __on_scene_list_new_button_value(self, value):
        if self.is_enabled():
            if value:
                if self._prepare_new_action():
                    song = self.song
                    view = song.view
                    try:
                        if liveobj_valid(view.highlighted_clip_slot.clip):
                            song.capture_and_insert_scene(Live.Song.CaptureMode.all_except_selected)
                        else:
                            view.selected_track.stop_all_clips(False)
                    except Live.Base.LimitationError:
                        self._handle_limitation_error_on_scene_creation()

                    self._view_selected_clip_detail()

    @listens("value")
    def __on_new_button_value(self, value):
        if self.is_enabled():
            if value:
                if self._prepare_new_action():
                    song = self.song
                    view = song.view
                    try:
                        selected_track = view.selected_track
                        selected_scene_index = list(song.scenes).index(view.selected_scene)
                        selected_track.stop_all_clips(False)
                        self._jump_to_next_slot(selected_track, selected_scene_index)
                    except Live.Base.LimitationError:
                        self._handle_limitation_error_on_scene_creation()

                    self._view_selected_clip_detail()

    def _prepare_new_action(self):
        song = self.song
        selected_track = song.view.selected_track
        if selected_track.can_be_armed:
            song.overdub = False
            return True
        return None

    def _has_clip(self, scene_or_track):
        return find_if(lambda x: liveobj_valid(x.clip), scene_or_track.clip_slots) is not None

    def _create_silent_scene(self, scene_index):
        song = self.song
        song.stop_all_clips(False)
        selected_scene = song.view.selected_scene
        if not selected_scene.is_empty:
            new_scene_index = list(song.scenes).index(selected_scene) + 1
            song.view.selected_scene = song.create_scene(new_scene_index)

    def _jump_to_next_slot(self, track, start_index):
        song = self.song
        new_scene_index = self._next_empty_slot(track, start_index)
        song.view.selected_scene = song.scenes[new_scene_index]

    def _stop_recording(self):
        song = self.song
        status = song.session_record_status
        was_recording = status != Live.Song.SessionRecordStatus.off or song.session_record
        if was_recording:
            song.session_record = False
        return was_recording

    def _start_recording(self):
        song = self.song
        song.session_record = True

    def _next_empty_slot(self, track, scene_index):
        song = self.song
        scene_count = len(song.scenes)
        while track.clip_slots[scene_index].has_clip:
            scene_index += 1
            if scene_index == scene_count:
                song.create_scene(scene_count)

        return scene_index

    def _view_selected_clip_detail(self):
        view = self.song.view
        if view.highlighted_clip_slot.clip:
            view.detail_clip = view.highlighted_clip_slot.clip
        if self._view_controller is not None:
            self._view_controller.show_view("Detail/Clip")

    def _reconnect_track_listeners(self):
        manager = self._track_subject_slots
        manager.disconnect()
        for track in self.song.tracks:
            if track.can_be_armed:
                manager.register_slot(track, self.update, "arm")
                manager.register_slot(track, self.update, "playing_slot_index")
                manager.register_slot(track, self.update, "fired_slot_index")

    @property
    def scene_list_mode(self):
        return self._scene_list_mode

    @scene_list_mode.setter
    def scene_list_mode(self, scene_list_mode):
        self._scene_list_mode = scene_list_mode
        self._update_new_button()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/elements/velocity_levels_element.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 857 bytes
from ...base import EventObject, listenable_property
from .proxy_element import ProxyElement


class NullVelocityLevels(EventObject):
    enabled = False
    target_note = -1
    target_channel = -1
    source_channel = -1
    notes = []

    @property
    def levels(self):
        return []

    @listenable_property
    def last_played_level(self):
        return -1.0


class VelocityLevelsElement(ProxyElement):
    def __init__(self, velocity_levels=None, *a, **k):
        super().__init__(proxied_object=velocity_levels, proxied_interface=(NullVelocityLevels()))

    def reset(self):
        self.notes = []
        self.source_channel = -1



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/note_editor.py
# Compiled at: 2023-12-18 13:26:15
# Size of source mod 2**32: 22487 bytes
from math import inf

from Live.Clip import MidiNoteSpecification

from ...base import EventObject, clamp, depends, in_range, listenable_property, listens
from ...live import liveobj_changed, liveobj_valid
from .. import Component
from ..controls import ButtonControl, control_matrix
from ..skin import LiveObjSkinEntry

PROPERTY_RANGE_KEYS = ("pitch", "start_time", "duration", "velocity")
RELATIVE_OFFSET = 0.24
DEFAULT_HEIGHT = 4
DEFAULT_WIDTH = 4
DEFAULT_VELOCITY = 100
DEFAULT_START_NOTE = 36
DEFAULT_STEP_TRANSLATION_CHANNEL = 1


def get_notes(clip, pitches, time, length, all_pitches=False):
    if len(pitches) > 1 or all_pitches:
        return clip.get_notes_extended(from_time=time, from_pitch=0, time_span=length, pitch_span=128)
    return clip.get_notes_extended(from_time=time, from_pitch=(pitches[0]), time_span=length, pitch_span=1)


def remove_notes(clip, pitches, time, length, all_pitches=False):
    if len(pitches) > 1 or all_pitches:
        clip.remove_notes_extended(from_time=time, from_pitch=0, time_span=length, pitch_span=128)
    else:
        clip.remove_notes_extended(from_time=time, from_pitch=(pitches[0]), time_span=length, pitch_span=1)


def property_ranges_for_notes(notes, start_time, property_ranges):
    for note in notes:
        note_values = [note.pitch, note.start_time, note.duration, note.velocity]
        note_values[1] -= start_time
        for key, value in zip(PROPERTY_RANGE_KEYS, note_values):
            if not property_ranges:
                property_ranges = {key: (inf, -inf) for key in PROPERTY_RANGE_KEYS}
            else:
                min_value, max_value = property_ranges[key]
                property_ranges[key] = (min(value, min_value), max(value, max_value))

    return property_ranges


class PitchProvider(EventObject):
    pitches = listenable_property.managed([DEFAULT_START_NOTE])
    is_polyphonic = listenable_property.managed(False)


class TimeStep:
    def __init__(self, start, length, *a, **k):
        (super().__init__)(*a, **k)
        self.start = start
        self.length = length

    @property
    def offset(self):
        return self.length * RELATIVE_OFFSET

    def left_boundary(self):
        return max(0, self.start - self.length * 0.2)

    def right_boundary(self):
        return max(0, self.start + self.length * 0.7)

    def filter_notes(self, notes):
        return list(filter(self.includes_note, notes))

    def clamp(self, time):
        return clamp(time, self.left_boundary(), self.right_boundary())

    def includes_note(self, note):
        return self.includes_time(note.start_time)

    def includes_time(self, time):
        return in_range(time - self.start + self.offset, 0, self.length)

    def connected_time_ranges(self):
        return [(self.start - self.offset, self.length)]


class StepButtonControl(ButtonControl):
    class State(ButtonControl.State):
        x = property(
            lambda self: self.coordinate[1],
        )
        y = property(
            lambda self: self.coordinate[0],
        )
        is_active = False


class NoteEditorComponent(Component):
    __events__ = ("clip_notes",)
    matrix = control_matrix(StepButtonControl)

    @depends(target_track=None, sequencer_clip=None, full_velocity=None, grid_resolution=None)
    def __init__(
        self,
        name="Note_Editor",
        translation_channel=DEFAULT_STEP_TRANSLATION_CHANNEL,
        full_velocity=None,
        target_track=None,
        sequencer_clip=None,
        grid_resolution=None,
        *a,
        **k,
    ):
        (super().__init__)(a, name=name, **k)
        self._full_velocity = full_velocity
        self._translation_channel = translation_channel
        self._page_time = 0.0
        self._pitches = [DEFAULT_START_NOTE]
        self._pitch_provider = PitchProvider()
        self._clip_notes = []
        self._clip = None
        self._sequencer_clip = sequencer_clip
        self._active_steps = []
        self._grid_resolution = grid_resolution
        self._NoteEditorComponent__on_resolution_changed.subject = self._grid_resolution
        self._nudge_offset = 0
        self._duration_offset = 0
        self._velocity_offset = 0
        self._pitch_offset = 0
        self._triplet_factor = 1.0
        self._is_triplet = False
        self._update_from_grid()
        self._target_track = target_track
        self._NoteEditorComponent__on_target_track_color_changed.subject = target_track
        self._NoteEditorComponent__on_sequencer_clip_changed.subject = sequencer_clip
        self._NoteEditorComponent__on_sequencer_clip_changed()

    @property
    def width(self):
        if self.matrix.width:
            return self.matrix.width
        return DEFAULT_WIDTH

    @property
    def height(self):
        if self.matrix.height:
            return self.matrix.height
        return DEFAULT_HEIGHT

    @property
    def step_count(self):
        return self.width * self.height

    @property
    def step_length(self):
        return self._grid_resolution.step_length

    @property
    def can_change_page(self):
        return not self._active_steps

    @property
    def page_time(self):
        return self._page_time

    @page_time.setter
    def page_time(self, time):
        if time != self._page_time:
            self._page_time = time
            self.page_time_changed()
            self._NoteEditorComponent__on_clip_notes_changed()

    @listenable_property
    def page_length(self):
        return self.step_count * self.step_length * self._triplet_factor

    @listenable_property
    def active_steps(self):
        return list(map(self._get_step_time_range, self._active_steps))

    @listenable_property
    def pitch_provider(self):
        return self._pitch_provider

    @pitch_provider.setter
    def pitch_provider(self, provider):
        self._pitch_provider = provider
        self._NoteEditorComponent__on_provider_polyphony_changed.subject = provider
        self._NoteEditorComponent__on_provided_pitches_changed.subject = provider
        self._NoteEditorComponent__on_provided_pitches_changed(provider.pitches if provider else [])
        self.notify_pitch_provider()

    def set_clip(self, clip):
        if liveobj_changed(clip, self._clip):
            self._clip = clip
            if self.can_change_page:
                self.page_time = 0.0
            self._NoteEditorComponent__on_clip_notes_changed.subject = clip
            self._NoteEditorComponent__on_clip_notes_changed()

    def set_pitches(self, pitches):
        if pitches != self._pitches:
            self._pitches = pitches
            enabled = self._can_edit()
            for button in self.matrix:
                button.enabled = enabled

            if enabled:
                self._NoteEditorComponent__on_clip_notes_changed()

    def set_pitch_offset(self, value):
        self._modify_note_property("_pitch_offset", value)

    def set_duration_offset(self, value):
        self._modify_note_property("_duration_offset", value)

    def set_velocity_offset(self, value):
        self._modify_note_property("_velocity_offset", value)

    def set_nudge_offset(self, value):
        self._modify_note_property("_nudge_offset", value)

    def set_matrix(self, matrix):
        self.matrix.set_control_element(matrix)
        for button in self.matrix:
            button.channel = self._translation_channel

        self._update_editor_matrix()

    def is_pitch_active(self, pitch):
        if self._has_clip():
            for step in self.active_steps:
                if get_notes(self._clip, (pitch,), step[0], step[1]):
                    return True

        return False

    def toggle_pitch_for_all_active_steps(self, pitch):
        for step in self._active_steps:
            time, _, pitches = self._get_notes_info_from_step(step)
            if pitch not in pitches:
                self._add_new_note_in_step(pitch, time)
            else:
                time_step = self._time_step(self._get_step_start_time(step))
                for time, length in time_step.connected_time_ranges():
                    remove_notes(self._clip, (pitch,), time, length)

    def can_nudge_by_offset(self, offset):
        for step in self.active_steps:
            notes = self._time_step(step[0]).filter_notes(self._clip_notes)
            time_step = self._time_step(step[0])
            for note in notes:
                new_start_time = time_step.clamp(note.start_time + offset)
                if new_start_time != note.start_time:
                    if in_range(new_start_time, time_step.left_boundary(), time_step.right_boundary()):
                        return True

        return False

    def get_note_property_ranges(self):
        property_ranges = {}
        if self._active_steps:
            for step in self._active_steps:
                start_time = self._get_step_start_time(step)
                property_ranges = property_ranges_for_notes(
                    self._time_step(start_time).filter_notes(self._clip_notes),
                    start_time,
                    property_ranges,
                )

        return property_ranges

    def page_time_changed(self):
        pass

    def _time_step(self, time):
        return TimeStep(time, self.step_length)

    def _can_edit(self):
        return len(self._pitches) != 0

    def _has_clip(self):
        return liveobj_valid(self._clip)

    def _can_press_or_release_step(self, step):
        width = self.width * self._triplet_factor if self._is_triplet else self.width
        return self._can_edit() and step.x < width and step.y < self.height

    def _get_step_start_time(self, step):
        step_time = step.x + step.y * self.width * self._triplet_factor
        return self._page_time + step_time * self.step_length

    def _get_step_time_range(self, step):
        time = self._get_step_start_time(step)
        return (time, time + self.step_length)

    def _get_notes_info_from_step(self, step):
        time = self._get_step_start_time(step)
        notes = self._time_step(time).filter_notes(self._clip_notes)
        pitches = [note.pitch for note in notes]
        return (time, notes, pitches)

    def _get_clip_notes_time_range(self):
        return (self._page_time - self._time_step(0).offset, self.page_length)

    @matrix.pressed
    def matrix(self, pad):
        self._on_pad_pressed(pad)

    def _on_pad_pressed(self, pad):
        if self.is_enabled():
            if not self._has_clip():
                self.set_clip(self._sequencer_clip.create_clip())
                self._update_from_grid()
            if self._can_press_or_release_step(pad):
                pad.is_active = True
                self._refresh_active_steps()

    @matrix.released_immediately
    def matrix(self, pad):
        self._on_pad_released(pad, can_add_or_remove=True)

    @matrix.released_delayed
    def matrix(self, pad):
        self._on_pad_released(pad)

    def _on_pad_released(self, pad, **k):
        if self.is_enabled():
            if self._can_press_or_release_step(pad):
                (self._on_release_step)(pad, **k)

    def _on_release_step(self, step, can_add_or_remove=False):
        if step.is_active:
            if can_add_or_remove:
                self._delete_notes_in_step(step)
                for pitch in self._pitches:
                    self._add_note_in_step(step, pitch)

        step.is_active = False
        self._refresh_active_steps()

    def _refresh_active_steps(self):
        active_steps = [s for s in self.matrix if s.is_active]
        if active_steps != self._active_steps:
            self._active_steps = active_steps
            self.notify_active_steps()

    def _release_active_steps(self):
        for step in self.matrix:
            self._on_release_step(step)

    def _add_note_in_step(self, step, pitch):
        if self._has_clip():
            time, notes, _ = self._get_notes_info_from_step(step)
            if not notes:
                self._add_new_note_in_step(pitch, time)

    def _add_new_note_in_step(self, pitch, time):
        velocity = 127 if self._full_velocity.enabled else DEFAULT_VELOCITY
        note = MidiNoteSpecification(
            pitch=pitch,
            start_time=time,
            duration=(self.step_length),
            velocity=velocity,
            mute=False,
        )
        self._clip.add_new_notes((note,))
        self._clip.deselect_all_notes()

    def _delete_notes_in_step(self, step):
        if self._has_clip():
            if self._can_edit():
                time_step = self._time_step(self._get_step_start_time(step))
                for time, length in time_step.connected_time_ranges():
                    remove_notes(self._clip, self._pitches, time, length, self._pitch_provider.is_polyphonic)

    def _modify_note_property(self, note_property, value):
        if self.is_enabled():
            setattr(self, note_property, getattr(self, note_property) + value)
            if self._active_steps:
                if self._has_clip():
                    if self._can_edit():
                        self._modify_step_notes(self._active_steps)
                        self._clip.apply_note_modifications(self._clip_notes)
                        self._update_editor_matrix()
                        self.notify_active_steps()
            self._reset_modifications()

    def _reset_modifications(self):
        self._pitch_offset = 0
        self._velocity_offset = 0
        self._duration_offset = 0
        self._nudge_offset = 0

    def _modify_step_notes(self, steps):
        notes = self._clip_notes
        for step in steps:
            time_step = self._time_step(self._get_step_start_time(step))
            for note in notes:
                self._modify_note(time_step, self._duration_offset, self._nudge_offset, note)

    def _modify_note(self, time_step, duration_offset, nudge_offset, note):
        if time_step.includes_time(note.start_time):
            note.start_time = time_step.clamp(note.start_time + nudge_offset)
            note.duration = max(time_step.length * 0.1, note.duration + duration_offset)
            note.velocity = clamp(note.velocity + self._velocity_offset, 1, 127)
            note.pitch = clamp(note.pitch + self._pitch_offset, 0, 127)

    def _update_from_grid(self):
        quantization, self._is_triplet = self._grid_resolution.clip_grid
        self._triplet_factor = 1.0 if (not self._is_triplet) else 0.75
        if self._has_clip():
            self._clip.view.grid_quantization = quantization
            self._clip.view.grid_is_triplet = self._is_triplet

    def update(self):
        super().update()
        self._update_editor_matrix()

    def _update_editor_matrix(self):
        if self.is_enabled():
            visible_steps = self._visible_steps()
            for index, button in enumerate(self.matrix):
                button.color = LiveObjSkinEntry(
                    self._get_color_for_step(index, visible_steps),
                    self._target_track.target_track,
                )

    def _get_color_for_step(self, index, visible_steps):
        color = "NoteEditor.StepDisabled"
        if self._has_clip() and index in visible_steps:
            notes = visible_steps[index].filter_notes(self._clip_notes)
            color = "NoteEditor.StepEmpty"
            if len(notes) > 0:
                color = "NoteEditor.StepMuted" if any(n.mute for n in notes) else "NoteEditor.StepFilled"
        else:
            if index in visible_steps:
                color = "NoteEditor.NoClip"
        return self._get_alternate_color_for_step(index, visible_steps) or color

    def _get_alternate_color_for_step(self, index, visible_steps):
        pass

    def _visible_steps(self):
        steps_per_page = self.step_count
        step_length = self.step_length
        indices = list(range(steps_per_page))
        if self._is_triplet:
            width = self.width
            columns_to_omit = [width - i - 1 for i in reversed(range(width // 3))]
            indices = [k for k in indices if k % width not in columns_to_omit]
        return {index: self._time_step(self._page_time + k * step_length) for k, index in []}

    @listens("notes")
    def __on_clip_notes_changed(self):
        self._clip_notes = []
        if self._has_clip():
            if self._can_edit():
                start, length = self._get_clip_notes_time_range()
                self._clip_notes = get_notes(
                    self._clip,
                    self._pitches,
                    start,
                    length,
                    self._pitch_provider.is_polyphonic,
                )
        self._update_editor_matrix()
        self.notify_clip_notes()

    @listens("clip")
    def __on_sequencer_clip_changed(self):
        self.set_clip(self._sequencer_clip.clip)

    @listens("target_track.color")
    def __on_target_track_color_changed(self):
        self._update_editor_matrix()

    @listens("pitches")
    def __on_provided_pitches_changed(self, pitches):
        self.set_pitches(pitches)

    @listens("is_polyphonic")
    def __on_provider_polyphony_changed(self, _):
        self._NoteEditorComponent__on_clip_notes_changed()

    @listens("index")
    def __on_resolution_changed(self, *_):
        self._release_active_steps()
        self._update_from_grid()
        self.notify_page_length()
        self._NoteEditorComponent__on_clip_notes_changed()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/clip_actions.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3021 bytes
import Live
from control_surface.control import ButtonControl

from ...base import duplicate_clip_loop, listens, liveobj_valid
from ...control_surface import Component


class ClipActionsComponent(Component):
    delete_button = ButtonControl(color="Action.Available")
    duplicate_button = ButtonControl(color="Action.Available")
    double_loop_button = ButtonControl(color="Action.Available")

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        view = self.song.view
        self._ClipActionsComponent__on_selected_scene_changed.subject = view
        self._ClipActionsComponent__on_selected_track_changed.subject = view
        self._ClipActionsComponent__on_selected_track_changed()

    @property
    def clip_slot(self):
        return self.song.view.highlighted_clip_slot

    @delete_button.released
    def delete_button(self, _):
        self.clip_slot.delete_clip()

    @duplicate_button.pressed
    def duplicate_button(self, _):
        clip_slot = self.clip_slot
        view = self.song.view
        track = view.selected_track
        try:
            source_index = list(track.clip_slots).index(clip_slot)
            target_index = track.duplicate_clip_slot(source_index)
            view.highlighted_clip_slot = track.clip_slots[target_index]
            view.detail_clip = view.highlighted_clip_slot.clip
            if clip_slot.clip.is_playing:
                view.highlighted_clip_slot.fire(force_legato=True, launch_quantization=(Live.Song.Quantization.q_no_q))
        except (Live.Base.LimitationError, RuntimeError):
            pass

    @double_loop_button.pressed
    def double_loop_button(self, _):
        duplicate_clip_loop(self.clip_slot.clip)

    @listens("selected_scene")
    def __on_selected_scene_changed(self):
        self._update_clip_slot()

    @listens("selected_track")
    def __on_selected_track_changed(self):
        self._ClipActionsComponent__on_input_routing_type_changed.subject = self.song.view.selected_track
        self._update_clip_slot()

    @listens("input_routing_type")
    def __on_input_routing_type_changed(self):
        self._update_action_buttons()

    @listens("has_clip")
    def __on_has_clip_changed(self):
        self._update_action_buttons()

    def _update_clip_slot(self):
        self._ClipActionsComponent__on_has_clip_changed.subject = self.song.view.highlighted_clip_slot
        self._ClipActionsComponent__on_has_clip_changed()

    def _update_action_buttons(self):
        can_perform_clip_action = self._can_perform_clip_action()
        self.delete_button.enabled = can_perform_clip_action
        self.duplicate_button.enabled = can_perform_clip_action
        self.double_loop_button.enabled = can_perform_clip_action and self.clip_slot.clip.is_midi_clip

    def _can_perform_clip_action(self):
        clip_slot = self.clip_slot
        return liveobj_valid(clip_slot) and clip_slot.has_clip



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/channel_strip.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 17646 bytes
from itertools import chain

from ...base import EventObject, listens, liveobj_valid, nop
from ..component import Component
from ..control import ButtonControl
from ..elements import DisplayDataSource


def release_control(control):
    if control is not None:
        control.release_parameter()


def reset_button(button):
    if button is not None:
        button.reset()


class ChannelStripComponent(Component):
    _active_instances = []

    def number_of_arms_pressed():
        result = 0
        for strip in ChannelStripComponent._active_instances:
            if strip.arm_button_pressed():
                result += 1

        return result

    number_of_arms_pressed = staticmethod(number_of_arms_pressed)

    def number_of_solos_pressed():
        result = 0
        for strip in ChannelStripComponent._active_instances:
            if strip.solo_button_pressed():
                result += 1

        return result

    number_of_solos_pressed = staticmethod(number_of_solos_pressed)
    empty_color = None
    select_button = ButtonControl()

    def __init__(self, *a, **k):
        (super().__init__)(self, *a, **k)
        ChannelStripComponent._active_instances.append(self)
        self._track = None
        self._send_controls = []
        self._pan_control = None
        self._volume_control = None
        self._mute_button = None
        self._solo_button = None
        self._arm_button = None
        self._shift_button = None
        self._crossfade_toggle = None
        self._shift_pressed = False
        self._solo_pressed = False
        self._arm_pressed = False
        self._invert_mute_feedback = False
        self._track_name_data_source = DisplayDataSource()
        self._update_track_name_data_source()
        self._empty_control_slots = self.register_disconnectable(EventObject())
        self._ChannelStripComponent__on_selected_track_changed.subject = self.song.view

        def make_property_slot(name, alias=None):
            alias = alias or name
            return self.register_slot(None, getattr(self, f"_on_{alias}_changed"), name)

        self._track_property_slots = [
            make_property_slot("mute"),
            make_property_slot("solo"),
            make_property_slot("arm"),
            make_property_slot("input_routing_type", "input_routing"),
            make_property_slot("name", "track_name"),
        ]
        self._mixer_device_property_slots = [
            make_property_slot("crossfade_assign", "cf_assign"),
            make_property_slot("sends"),
        ]

        def make_button_slot(name):
            return self.register_slot(None, getattr(self, f"_{name}_value"), "value")

        self._mute_button_slot = make_button_slot("mute")
        self._solo_button_slot = make_button_slot("solo")
        self._arm_button_slot = make_button_slot("arm")
        self._shift_button_slot = make_button_slot("shift")
        self._crossfade_toggle_slot = make_button_slot("crossfade_toggle")

    def disconnect(self):
        ChannelStripComponent._active_instances.remove(self)
        for button in [
            self._mute_button,
            self._solo_button,
            self._arm_button,
            self._shift_button,
            self._crossfade_toggle,
        ]:
            reset_button(button)

        for control in self._all_controls():
            release_control(control)

        self._track_name_data_source.set_display_string("")
        self._mute_button = None
        self._solo_button = None
        self._arm_button = None
        self._shift_button = None
        self._crossfade_toggle = None
        self._track_name_data_source = None
        self._pan_control = None
        self._volume_control = None
        self._send_controls = None
        self._track = None
        super().disconnect()

    def set_track(self, track):
        for control in self._all_controls():
            release_control(control)

        self._track = track
        for slot in self._track_property_slots:
            slot.subject = track

        for slot in self._mixer_device_property_slots:
            slot.subject = track.mixer_device if liveobj_valid(track) else None

        if liveobj_valid(self._track):
            for button in (self._mute_button, self._solo_button, self._arm_button, self._crossfade_toggle):
                if button is not None:
                    button.set_light(False)

        self.select_button.enabled = bool(self._track)
        self._update_track_name_data_source()
        self.update()

    def reset_button_on_exchange(self, button):
        reset_button(button)

    def _update_track_name_data_source(self):
        self._track_name_data_source.set_display_string(self._track.name if liveobj_valid(self._track) else " - ")

    def set_send_controls(self, controls):
        for control in list(self._send_controls or []):
            release_control(control)

        self._send_controls = controls
        self.update()

    def set_pan_control(self, control):
        if control != self._pan_control:
            release_control(self._pan_control)
            self._pan_control = control
            self.update()

    def set_volume_control(self, control):
        if control != self._volume_control:
            release_control(self._volume_control)
            self._volume_control = control
            self.update()

    def set_select_button(self, button):
        self.select_button.set_control_element(button)
        self.update()

    def set_mute_button(self, button):
        if button != self._mute_button:
            self.reset_button_on_exchange(self._mute_button)
            self._mute_button = button
            self._mute_button_slot.subject = button
            self.update()

    def set_solo_button(self, button):
        if button != self._solo_button:
            self.reset_button_on_exchange(self._solo_button)
            self._solo_pressed = False
            self._solo_button = button
            self._solo_button_slot.subject = button
            self.update()

    def set_arm_button(self, button):
        if button != self._arm_button:
            self.reset_button_on_exchange(self._arm_button)
            self._arm_pressed = False
            self._arm_button = button
            self._arm_button_slot.subject = button
            self.update()

    def set_shift_button(self, button):
        if button != self._shift_button:
            self.reset_button_on_exchange(self._shift_button)
            self._shift_button = button
            self._shift_button_slot.subject = button
            self.update()

    def set_crossfade_toggle(self, button):
        if button != self._crossfade_toggle:
            self.reset_button_on_exchange(self._crossfade_toggle)
            self._crossfade_toggle = button
            self._crossfade_toggle_slot.subject = button
            self.update()

    def set_invert_mute_feedback(self, invert_feedback):
        if invert_feedback != self._invert_mute_feedback:
            self._invert_mute_feedback = invert_feedback
            self.update()

    @listens("selected_track")
    def __on_selected_track_changed(self):
        self._update_select_button()

    def _update_select_button(self):
        if liveobj_valid(self._track) or self.empty_color is None:
            if self.song.view.selected_track == self._track:
                self.select_button.color = "DefaultButton.On"
            else:
                self.select_button.color = "DefaultButton.Off"
        else:
            self.select_button.color = self.empty_color

    def solo_button_pressed(self):
        return self._solo_pressed

    def arm_button_pressed(self):
        return self._arm_pressed

    def track_name_data_source(self):
        return self._track_name_data_source

    @property
    def track(self):
        return self._track

    def _connect_parameters(self):
        if self._pan_control is not None:
            self._pan_control.connect_to(self._track.mixer_device.panning)
        if self._volume_control is not None:
            self._volume_control.connect_to(self._track.mixer_device.volume)
        if self._send_controls is not None:
            index = 0
            for send_control in self._send_controls:
                if send_control is not None:
                    if index < len(self._track.mixer_device.sends):
                        send_control.connect_to(self._track.mixer_device.sends[index])
                    else:
                        send_control.release_parameter()
                        self._empty_control_slots.register_slot(send_control, nop, "value")
                index += 1

    def _all_controls(self):
        return [self._pan_control, self._volume_control, *list(self._send_controls or [])]

    def _disconnect_parameters(self):
        for control in self._all_controls():
            release_control(control)
            self._empty_control_slots.register_slot(control, nop, "value")

    def update(self):
        super().update()
        if self.is_enabled():
            self._empty_control_slots.disconnect()
            if liveobj_valid(self._track):
                self._connect_parameters()
            else:
                self._disconnect_parameters()
            self._ChannelStripComponent__on_selected_track_changed()
            self._on_mute_changed()
            self._on_solo_changed()
            self._on_arm_changed()
            self._on_cf_assign_changed()
        else:
            self._disconnect_parameters()

    @select_button.pressed
    def select_button(self, button):
        self._on_select_button_pressed(button)

    def _on_select_button_pressed(self, button):
        if liveobj_valid(self._track):
            if self.song.view.selected_track != self._track:
                self.song.view.selected_track = self._track

    @select_button.pressed_delayed
    def select_button(self, button):
        self._on_select_button_pressed_delayed(button)

    def _on_select_button_pressed_delayed(self, button):
        pass

    @select_button.released
    def select_button(self, button):
        self._on_select_button_released(button)

    def _on_select_button_released(self, button):
        pass

    @select_button.double_clicked
    def select_button(self, button):
        self._on_select_button_double_clicked(button)

    def _on_select_button_double_clicked(self, button):
        pass

    def _mute_value(self, value):
        if self.is_enabled():
            if liveobj_valid(self._track):
                if self._track != self.song.master_track:
                    if not self._mute_button.is_momentary() or value != 0:
                        self._track.mute = not self._track.mute

    def update_solo_state():
        pass

    def _solo_value(self, value):
        if self.is_enabled():
            if not liveobj_valid(self._track) or self._track != self.song.master_track:
                self._solo_pressed = value != 0 and self._solo_button.is_momentary()
                if not (value != 0 or self._solo_button.is_momentary()):
                    expected_solos_pressed = 1 if self._solo_pressed else 0
                    solo_exclusive = self.song.exclusive_solo != self._shift_pressed and (
                        not self._solo_button.is_momentary()
                        or ChannelStripComponent.number_of_solos_pressed() == expected_solos_pressed
                    )
                    new_value = not self._track.solo
                    respect_multi_selection = self._track.is_part_of_selection
                    for track in chain(self.song.tracks, self.song.return_tracks):
                        self.update_solo_state(solo_exclusive, new_value, respect_multi_selection, track)

    def _arm_value():
        pass

    def _shift_value(self, value):
        self._shift_pressed = value != 0

    def _crossfade_toggle_value(self, value):
        if self.is_enabled():
            if liveobj_valid(self._track):
                if not (value != 0 or self._crossfade_toggle.is_momentary()):
                    self._track.mixer_device.crossfade_assign = (self._track.mixer_device.crossfade_assign - 1) % len(
                        self._track.mixer_device.crossfade_assignments.values,
                    )

    def _on_sends_changed(self):
        if self.is_enabled():
            self.update()

    def _on_mute_changed(self):
        if (
            not self.is_enabled()
            or self._mute_button is not None
            and liveobj_valid(self._track)
            or self.empty_color is None
        ):
            if (
                self._track in chain(self.song.tracks, self.song.return_tracks)
                and self._track.mute != self._invert_mute_feedback
            ):
                self._mute_button.set_light("Mixer.MuteOff")
            else:
                self._mute_button.set_light("Mixer.MuteOn")
        else:
            self._mute_button.set_light(self.empty_color)

    def _on_solo_changed(self):
        if (
            not self.is_enabled()
            or self._solo_button is not None
            and liveobj_valid(self._track)
            or self.empty_color is None
        ):
            if self._track in chain(self.song.tracks, self.song.return_tracks) and self._track.solo:
                self._solo_button.set_light("Mixer.SoloOn")
            else:
                self._solo_button.set_light("Mixer.SoloOff")
        else:
            self._solo_button.set_light(self.empty_color)

    def _on_arm_changed(self):
        if (
            not self.is_enabled()
            or self._arm_button is not None
            and liveobj_valid(self._track)
            or self.empty_color is None
        ):
            if self._track in self.song.tracks and self._track.can_be_armed and self._track.arm:
                self._arm_button.set_light("Mixer.ArmOn")
            else:
                self._arm_button.set_light("Mixer.ArmOff")
        else:
            self._arm_button.set_light(self.empty_color)

    def _on_track_name_changed(self):
        if liveobj_valid(self._track):
            self._update_track_name_data_source()

    def _on_cf_assign_changed(self):
        if self.is_enabled():
            if self._crossfade_toggle is not None:
                if (
                    liveobj_valid(self._track)
                    and self._track in chain(self.song.tracks, self.song.return_tracks)
                    and self._track.mixer_device.crossfade_assign != 1
                ):
                    self._crossfade_toggle.set_light(True)
                else:
                    self._crossfade_toggle.set_light(False)

    def _on_input_routing_changed(self):
        if self.is_enabled():
            self._on_arm_changed()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/SessionZoomingComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 14612 bytes
from past.utils import old_div

from .CompoundComponent import CompoundComponent
from .ScrollComponent import ScrollComponent
from .SubjectSlot import subject_slot, subject_slot_group
from .Util import in_range


class SessionZoomingComponent(CompoundComponent):
    def __init__(self, session=None, enable_skinning=False, *a, **k):
        (super().__init__)(*a, **k)
        self._buttons = None
        self._scene_bank_buttons = None
        self._scene_bank_button_slots = self.register_slot_manager()
        self._scene_bank_index = 0
        self._empty_value = 0
        self._stopped_value = 100
        self._playing_value = 127
        self._selected_value = 64
        self._session, self._vertical_scroll, self._horizontal_scroll = self.register_components(
            session,
            ScrollComponent(),
            ScrollComponent(),
        )
        self._vertical_scroll.can_scroll_up = self._can_scroll_up
        self._vertical_scroll.can_scroll_down = self._can_scroll_down
        self._vertical_scroll.scroll_up = self._scroll_up
        self._vertical_scroll.scroll_down = self._scroll_down
        self._horizontal_scroll.can_scroll_up = self._can_scroll_left
        self._horizontal_scroll.can_scroll_down = self._can_scroll_right
        self._horizontal_scroll.scroll_up = self._scroll_left
        self._horizontal_scroll.scroll_down = self._scroll_right
        self.register_slot(self._session, self._on_session_offset_changes, "offset")
        if enable_skinning:
            self._enable_skinning()

    def _enable_skinning(self):
        self.set_stopped_value("Zooming.Stopped")
        self.set_selected_value("Zooming.Selected")
        self.set_playing_value("Zooming.Playing")
        self.set_empty_value("Zooming.Empty")

    def on_scene_list_changed(self):
        self.update()

    def on_enabled_changed(self):
        super().on_enabled_changed()
        self._session.set_show_highlight(self.is_enabled())

    def set_button_matrix(self, buttons):
        if buttons:
            buttons.reset()
        self._buttons = buttons
        self._on_matrix_value.subject = self._buttons
        self.update()
        self._on_session_offset_changes()

    def set_nav_buttons(self, up, down, left, right):
        self.set_nav_up_button(up)
        self.set_nav_down_button(down)
        self.set_nav_left_button(left)
        self.set_nav_right_button(right)
        self.update()

    def set_nav_up_button(self, button):
        self._vertical_scroll.set_scroll_up_button(button)

    def set_nav_down_button(self, button):
        self._vertical_scroll.set_scroll_down_button(button)

    def set_nav_left_button(self, button):
        self._horizontal_scroll.set_scroll_up_button(button)

    def set_nav_right_button(self, button):
        self._horizontal_scroll.set_scroll_down_button(button)

    def set_scene_bank_buttons(self, buttons):
        self._scene_bank_buttons = buttons
        self._on_scene_bank_value.replace_subjects(buttons or [])
        self.update()

    def set_empty_value(self, value):
        self._empty_value = value

    def set_playing_value(self, value):
        self._playing_value = value

    def set_stopped_value(self, value):
        self._stopped_value = value

    def set_selected_value(self, value):
        self._selected_value = value

    def update(self):
        super().update()
        if self._allow_updates:
            if self.is_enabled():
                self._update_matrix_buttons()
                self._update_scene_bank_buttens()
        else:
            self._update_requests += 1

    def _update_matrix_buttons(self):
        if self._buttons is not None:
            tracks = self._session.tracks_to_use()
            scenes = self.song().scenes
            slots_registry = [None for index in range(len(scenes))]
            width = self._session.width()
            height = self._session.height()
            for x in range(self._buttons.width()):
                for y in range(self._buttons.height()):
                    value_to_send = self._empty_value
                    scene_bank_offset = self._scene_bank_index * self._buttons.height() * height
                    track_offset = x * width
                    scene_offset = y * height + scene_bank_offset
                    if track_offset in range(len(tracks)):
                        if scene_offset in range(len(scenes)):
                            value_to_send = self._stopped_value
                            if self._session.track_offset() in range(
                                width * (x - 1) + 1,
                                width * (x + 1),
                            ) and self._session.scene_offset() - scene_bank_offset in range(
                                height * (y - 1) + 1,
                                height * (y + 1),
                            ):
                                value_to_send = self._selected_value
                            else:
                                playing = False
                                for track in range(track_offset, track_offset + width):
                                    for scene in range(scene_offset, scene_offset + height):
                                        if track in range(len(tracks)):
                                            if scene in range(len(scenes)):
                                                if slots_registry[scene] is None:
                                                    slots_registry[scene] = scenes[scene].clip_slots
                                                else:
                                                    slot = (
                                                        slots_registry[scene][track]
                                                        if len(slots_registry[scene]) > track
                                                        else None
                                                    )
                                                if slot is not None:
                                                    if slot.has_clip:
                                                        if slot.clip.is_playing:
                                                            value_to_send = self._playing_value
                                                            playing = True
                                                            break

                                    if playing:
                                        break

                    if in_range(value_to_send, 0, 128):
                        self._buttons.send_value(x, y, value_to_send)
                    else:
                        self._buttons.set_light(x, y, value_to_send)

    def _update_scene_bank_buttens(self):
        if self._scene_bank_buttons is not None:
            for index, button in enumerate(self._scene_bank_buttons):
                if button:
                    button.set_light(index == self._scene_bank_index)

    def _on_session_offset_changes(self):
        if self._buttons:
            self._scene_bank_index = int(
                old_div(old_div(self._session.scene_offset(), self._session.height()), self._buttons.height()),
            )
        self.update()

    @subject_slot("value")
    def _on_matrix_value(self, value, x, y, is_momentary):
        if self.is_enabled():
            if not (value != 0 or is_momentary):
                track_offset = x * self._session.width()
                scene_offset = (y + self._scene_bank_index * self._buttons.height()) * self._session.height()
                if track_offset in range(len(self._session.tracks_to_use())):
                    if scene_offset in range(len(self.song().scenes)):
                        self._session.set_offsets(track_offset, scene_offset)

    @subject_slot_group("value")
    def _on_scene_bank_value(self, value, sender):
        if not self.is_enabled() or self._buttons:
            if not (value != 0 or sender.is_momentary()):
                button_offset = list(self._scene_bank_buttons).index(sender)
                scene_offset = button_offset * self._buttons.height() * self._session.height()
                if scene_offset in range(len(self.song().scenes)):
                    self._scene_bank_index = button_offset
                    self.update()

    def _can_scroll_up(self):
        return self._session._can_bank_up()

    def _can_scroll_down(self):
        return self._session._can_bank_down()

    def _can_scroll_left(self):
        return self._session._can_bank_left()

    def _can_scroll_right(self):
        return self._session._can_bank_right()

    def _scroll_up(self):
        height = self._session.height()
        track_offset = self._session.track_offset()
        scene_offset = self._session.scene_offset()
        if scene_offset > 0:
            new_scene_offset = scene_offset
            if scene_offset % height > 0:
                new_scene_offset -= scene_offset % height
            else:
                new_scene_offset = max(0, scene_offset - height)
            self._session.set_offsets(track_offset, new_scene_offset)

    def _scroll_down(self):
        height = self._session.height()
        track_offset = self._session.track_offset()
        scene_offset = self._session.scene_offset()
        new_scene_offset = scene_offset + height - scene_offset % height
        self._session.set_offsets(track_offset, new_scene_offset)

    def _scroll_left(self):
        width = self._session.width()
        track_offset = self._session.track_offset()
        scene_offset = self._session.scene_offset()
        if track_offset > 0:
            new_track_offset = track_offset
            if track_offset % width > 0:
                new_track_offset -= track_offset % width
            else:
                new_track_offset = max(0, track_offset - width)
            self._session.set_offsets(new_track_offset, scene_offset)

    def _scroll_right(self):
        width = self._session.width()
        track_offset = self._session.track_offset()
        scene_offset = self._session.scene_offset()
        new_track_offset = track_offset + width - track_offset % width
        self._session.set_offsets(new_track_offset, scene_offset)


class DeprecatedSessionZoomingComponent(SessionZoomingComponent):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._zoom_button = None
        self._is_zoomed_out = False

    def set_button_matrix(self, buttons):
        self._buttons = buttons
        self._on_matrix_value.subject = self._buttons
        self.update()
        self._on_session_offset_changes()

    def set_zoom_button(self, button):
        if button != self._zoom_button:
            self._zoom_button = button
            self._on_zoom_value.subject = button
            self._is_zoomed_out = False
            self.update()

    def _session_set_enabled(self, is_enabled):
        self._session.set_enabled(is_enabled)

    @subject_slot("value")
    def _on_zoom_value(self, value):
        if self.is_enabled():
            if self._zoom_button.is_momentary():
                self._is_zoomed_out = value > 0
            else:
                self._is_zoomed_out = not self._is_zoomed_out
            if self._is_zoomed_out and self._buttons:
                self._scene_bank_index = int(
                    old_div(old_div(self._session.scene_offset(), self._session.height()), self._buttons.height()),
                )
            else:
                self._scene_bank_index = 0
            self._session_set_enabled(not self._is_zoomed_out)
            self.update()

    def update(self):
        if self._allow_updates:
            self._session_set_enabled(not self._is_zoomed_out)
        super().update()

    def _update_matrix_buttons(self):
        if self._is_zoomed_out:
            super()._update_matrix_buttons()

    def _update_scene_bank_buttens(self):
        if self._is_zoomed_out:
            super()._update_scene_bank_buttens()

    def _on_session_offset_changes(self):
        if self._is_zoomed_out:
            super()._on_session_offset_changes()

    @subject_slot("value")
    def _on_matrix_value(self, *a, **k):
        if self._is_zoomed_out:
            (super()._on_matrix_value)(*a, **k)

    @subject_slot_group("value")
    def _on_scene_bank_value(self, *a, **k):
        if self._is_zoomed_out:
            (super()._on_scene_bank_value)(*a, **k)

    def _scroll_up(self):
        if self._is_zoomed_out:
            super()._scroll_up()

    def _scroll_down(self):
        if self._is_zoomed_out:
            super()._scroll_down()

    def _scroll_left(self):
        if self._is_zoomed_out:
            super()._scroll_left()

    def _scroll_right(self):
        if self._is_zoomed_out:
            super()._scroll_right()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/util.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1957 bytes
from typing import List

from ...base import BooleanContext

updating_display = BooleanContext()


def auto_break_lines(text: str, max_width: int, max_lines: int, pad_lines: bool = True) -> List[str]:
    words = [truncate_with_ellipses(word[:max_width]) if len(word) > max_width else word for word in text.split(" ")]
    lines = [""]
    current_width = 0
    for word in words:
        if current_width + len(word) <= max_width:
            str_to_append = f" {word}" if current_width != 0 else word
            current_width += len(str_to_append)
            lines[-1] += str_to_append
        else:
            lines.append(word)
            current_width = len(word)

    if len(lines) > max_lines:
        lines = lines[:max_lines]
        lines[-1] = truncate_with_ellipses(lines[-1])
    if pad_lines:
        if len(lines) < max_lines:
            lines += [""] * (max_lines - len(lines))
    return lines


def truncate_with_ellipses(string: str):
    if len(string) > 3:
        return string[: len(string) - 3] + "..."
    return "." * len(string)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/EncoderElement.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 7712 bytes
import Live
from past.utils import old_div

from .ComboElement import WrapperElement
from .CompoundElement import CompoundElement
from .InputControlElement import MIDI_CC_TYPE, InputControlElement, InputSignal
from .SubjectSlot import SubjectEvent, subject_slot
from .Util import const, nop


def _not_implemented(value):
    raise NotImplementedError


_map_modes = map_modes = Live.MidiMap.MapMode
ENCODER_VALUE_NORMALIZER = {
    _map_modes.relative_smooth_two_compliment: lambda v: v if v <= 64 else v - 128,
    _map_modes.relative_smooth_signed_bit: lambda v: v if v <= 64 else 64 - v,
    _map_modes.relative_smooth_binary_offset: lambda v: v - 64,
}
MAX_14_BIT_CC = 95


class EncoderElement(InputControlElement):
    class ProxiedInterface(InputControlElement.ProxiedInterface):
        normalize_value = nop

    __subject_events__ = (SubjectEvent(name="normalized_value", signal=InputSignal),)
    encoder_sensitivity = 1.0

    def __init__(self, msg_type, channel, identifier, map_mode, encoder_sensitivity=None, *a, **k):
        (super().__init__)(msg_type, channel, identifier, *a, **k)
        if encoder_sensitivity is not None:
            self.encoder_sensitivity = encoder_sensitivity
        if map_mode is _map_modes.absolute_14_bit and identifier > MAX_14_BIT_CC:
            self._EncoderElement__map_mode = _map_modes.absolute
        else:
            self._EncoderElement__map_mode = map_mode
        self._EncoderElement__value_normalizer = ENCODER_VALUE_NORMALIZER.get(map_mode, _not_implemented)

    def message_map_mode(self):
        return self._EncoderElement__map_mode

    def relative_value_to_delta(self, value):
        return self._EncoderElement__value_normalizer(value)

    def normalize_value(self, value):
        return old_div(self.relative_value_to_delta(value), 64.0) * self.encoder_sensitivity

    def notify_value(self, value):
        super().notify_value(value)
        if self.normalized_value_listener_count():
            self.notify_normalized_value(self.normalize_value(value))


class TouchEncoderElementBase(EncoderElement):
    class ProxiedInterface(EncoderElement.ProxiedInterface):
        is_pressed = const(False)
        add_touch_value_listener = nop
        remove_touch_value_listener = nop
        touch_value_has_listener = nop

    __subject_events__ = ("touch_value",)

    def is_pressed(self):
        raise NotImplementedError


class TouchEncoderElement(CompoundElement, TouchEncoderElementBase):
    def __init__(self, channel=0, identifier=0, map_mode=_map_modes.absolute, touch_element=None, *a, **k):
        (super().__init__)(MIDI_CC_TYPE, channel, identifier, map_mode, *a, **k)
        self._touch_element = self.register_control_element(touch_element)

    def add_touch_value_listener(self, *a, **k):
        if self.value_listener_count() == 0:
            self.request_listen_nested_control_elements()
        (super().add_touch_value_listener)(*a, **k)

    def remove_touch_value_listener(self, *a, **k):
        (super().remove_touch_value_listener)(*a, **k)
        if self.value_listener_count() == 0:
            self.unrequest_listen_nested_control_elements()

    def on_nested_control_element_value(self, value, control):
        self.notify_touch_value(value)

    def is_pressed(self):
        return self.owns_control_element(self._touch_element) and self._touch_element.is_pressed()

    def on_nested_control_element_received(self, control):
        pass

    def on_nested_control_element_lost(self, control):
        pass


class FineGrainWithModifierEncoderElement(WrapperElement):
    def __init__(self, encoder=None, modifier=None, modified_sensitivity=0.1, default_sensitivity=None, *a, **k):
        (super().__init__)(a, wrapped_control=encoder, **k)
        self._modified_sensitivity = modified_sensitivity
        self._default_sensitivity = default_sensitivity or self.wrapped_control.mapping_sensitivity
        self._normalized_value_listeners = []
        self._modifier = modifier
        self.register_control_elements(modifier, encoder)
        self._on_nested_control_element_value.add_subject(self._modifier)

    def add_normalized_value_listener(self, listener):
        self._normalized_value_listeners.append(listener)
        if len(self._normalized_value_listeners) == 1:
            self._enforce_control_invariant()

    def remove_normalized_value_listener(self, listener):
        self._normalized_value_listeners.remove(listener)
        if len(self._normalized_value_listeners) == 0:
            self._enforce_control_invariant()

    def normalized_value_has_listener(self, listener):
        return listener in self._normalized_value_listeners

    @subject_slot("normalized_value")
    def __on_normalized_value(self, value):
        for listener in self._normalized_value_listeners:
            listener(value)

    def on_nested_control_element_received(self, control):
        super().on_nested_control_element_received(control)
        self._enforce_control_invariant()

    def on_nested_control_element_lost(self, control):
        super().on_nested_control_element_lost(control)
        self._enforce_control_invariant()

    def on_nested_control_element_value(self, value, control):
        if control == self._modifier:
            self._enforce_control_invariant()
        else:
            super().on_nested_control_element_value(value, control)

    def _enforce_control_invariant(self):
        if self.owns_control_element(self._wrapped_control):
            if self._modifier.is_pressed():
                self.wrapped_control.mapping_sensitivity = self._modified_sensitivity
            else:
                self.wrapped_control.mapping_sensitivity = self._default_sensitivity
        should_listen = self.owns_control_element(self._wrapped_control) and len(self._normalized_value_listeners) > 0
        self._FineGrainWithModifierEncoderElement__on_normalized_value.subject = (
            self._wrapped_control if should_listen else None
        )

    def set_sensitivities(self, default, modified):
        self._default_sensitivity = default
        self._modified_sensitivity = modified
        self._enforce_control_invariant()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/legacy_bank_definitions.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1948 bytes
from _Generic.Devices import BANK_NAME_DICT, DEVICE_BOB_DICT, DEVICE_DICT

from ableton.v2.base.collection import IndexedDict

from ..base import memoize
from . import BANK_PARAMETERS_KEY

BEST_OF_PARAMETERS_KEY = "Best of Parameters"


@memoize
def banked():
    return {
        device_name: IndexedDict(
            [
                (bank_name, {BANK_PARAMETERS_KEY: parameters})
                for bank_name, parameters in zip(bank_names, DEVICE_DICT[device_name])
            ],
        )
        for device_name, bank_names in BANK_NAME_DICT.items()
    }


@memoize
def best_of_banks():
    return {
        device_name: IndexedDict([(BEST_OF_PARAMETERS_KEY, {BANK_PARAMETERS_KEY: parameters[0]})])
        for device_name, parameters in DEVICE_BOB_DICT.items()
    }



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/elements/button_matrix.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1431 bytes
from ableton.v2.base import slicer, to_slice
from ableton.v2.control_surface.elements import ButtonMatrixElement as ButtonMatrixElementBase

from ...base import lazy_attribute, recursive_map
from ..display import Renderable


class ButtonMatrixElement(ButtonMatrixElementBase, Renderable):
    @property
    @slicer(2)
    def submatrix(self, col_slice, row_slice):
        col_slice = to_slice(col_slice)
        row_slice = to_slice(row_slice)
        rows = [row[col_slice] for row in self._orig_buttons[row_slice]]
        return ButtonMatrixElement(rows=rows)

    @lazy_attribute
    def renderable_state(self):
        if not any(isinstance(button, Renderable) for row in self._orig_buttons for button in row):
            return None
        matrix = recursive_map(
            lambda button: button.renderable_state if isinstance(button, Renderable) else None,
            self._orig_buttons,
        )
        if len(matrix) == 1:
            return matrix[0]
        return matrix



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/banking_util.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3892 bytes
from copy import deepcopy
from math import ceil

from past.utils import old_div

from ..base import liveobj_valid

MX_MAIN_BANK_INDEX = -1
BANK_FORMAT = "Bank %d"
BANK_PARAMETERS_KEY = "Parameters"
BANK_MAIN_KEY = "Main"


def has_bank_count(device):
    if liveobj_valid(device):
        try:
            num_banks = device.get_bank_count()
            return num_banks > 0
        except (AttributeError, RuntimeError):
            pass

        return False
    return None


def has_main_bank(device, definitions):
    if has_bank_count(device):
        try:
            main_bank = device.get_bank_parameters(MX_MAIN_BANK_INDEX)
            return bool(main_bank)
        except (AttributeError, RuntimeError):
            return False

    else:
        return BANK_MAIN_KEY in definitions.get(device.class_name, {})


def has_bank_names(device, definitions):
    if has_bank_count(device):
        try:
            name = device.get_bank_name(0)
            return bool(name)
        except (AttributeError, RuntimeError):
            return False

    else:
        return bool(definitions.get(device.class_name, {}).keys())


def all_parameters(device):
    if liveobj_valid(device):
        return list(device.parameters[1:])
    return []


def device_bank_count(device, bank_size=8, definition=None, definitions=None):
    count = 0
    if liveobj_valid(device):
        definition = definition or definitions.get(device.class_name, {})
        if has_bank_count(device):
            count = device.get_bank_count() + int(has_main_bank(device, definitions))
        else:
            if definition:
                count = len(definition.keys())
            else:
                count = int(ceil(old_div(float(len(all_parameters(device))), bank_size)))
    return count


def device_bank_definition(device, definitions):
    original_definition = definitions.get(device.class_name, None)
    return deepcopy(original_definition) if original_definition is not None else None


def device_bank_names(device, bank_size=8, definitions=None):
    names = []
    if liveobj_valid(device):
        class_name = device.class_name
        if class_name in definitions:
            names = definitions[class_name].keys()
        else:
            if has_bank_count(device) and has_bank_names(device, definitions):
                offset = int(has_main_bank(device, definitions))
                names = [
                    device.get_bank_name(index - offset)
                    for index in range(device_bank_count(device, definitions=definitions))
                ]
                if has_main_bank(device, definitions):
                    names[0] = names[0] or BANK_MAIN_KEY
            else:
                bank_count = device_bank_count(device, bank_size=bank_size, definitions=definitions)
                names = [BANK_FORMAT % (index + 1) for index in range(bank_count)]
    return names


class BankingInfo:
    def __init__(self, bank_definitions):
        self._bank_definitions = bank_definitions

    def has_bank_count(self, device):
        return has_bank_count(device)

    def has_main_bank(self, device):
        return has_main_bank(device, definitions=(self._bank_definitions))

    def has_bank_names(self, device):
        return has_bank_names(device, definitions=(self._bank_definitions))

    def device_bank_count(self, device, **k):
        return device_bank_count(device, definitions=self._bank_definitions, **k)

    def device_bank_definition(self, device):
        return device_bank_definition(device, definitions=(self._bank_definitions))

    def device_bank_names(self, device, **k):
        return device_bank_names(device, definitions=self._bank_definitions, **k)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/slide.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4111 bytes
from ...base import EventObject, clamp, listens
from ..component import Component
from .scroll import Scrollable, ScrollComponent


class Slideable(EventObject):
    __events__ = ("page_offset", "page_length", "position", "position_count", "contents")

    def contents_range(self, pmin, pmax):
        pos_count = self.position_count
        first_pos = max(int(pmin), 0)
        last_pos = min(int(pmax), pos_count)
        return list(range(first_pos, last_pos))

    def contents(self, position):
        return False

    @property
    def position_count(self):
        raise NotImplementedError

    @property
    def position(self):
        raise NotImplementedError

    @property
    def page_offset(self):
        raise NotImplementedError

    @property
    def page_length(self):
        raise NotImplementedError


class SlideComponent(Component, Scrollable):
    def __init__(self, slideable=None, *a, **k):
        (super().__init__)(*a, **k)
        slideable = slideable or self
        self._slideable = slideable
        self._position_scroll = ScrollComponent(parent=self)
        self._page_scroll = ScrollComponent(parent=self)
        self._position_scroll.scrollable = self
        self._page_scroll.can_scroll_up = self.can_scroll_page_up
        self._page_scroll.can_scroll_down = self.can_scroll_page_down
        self._page_scroll.scroll_down = self.scroll_page_down
        self._page_scroll.scroll_up = self.scroll_page_up
        self._SlideComponent__on_position_changed.subject = slideable

    def set_scroll_up_button(self, button):
        self._position_scroll.set_scroll_up_button(button)

    def set_scroll_down_button(self, button):
        self._position_scroll.set_scroll_down_button(button)

    def set_scroll_page_up_button(self, button):
        self._page_scroll.set_scroll_up_button(button)

    def set_scroll_page_down_button(self, button):
        self._page_scroll.set_scroll_down_button(button)

    def scroll_page_up(self):
        self._scroll_page(1)

    def scroll_page_down(self):
        self._scroll_page(-1)

    def scroll_up(self):
        self._scroll_position(1)

    def scroll_down(self):
        self._scroll_position(-1)

    def can_scroll_page_up(self):
        model = self._slideable
        return model.position < model.position_count - model.page_length

    def can_scroll_page_down(self):
        return self._slideable.position > 0

    def can_scroll_up(self):
        return self.can_scroll_page_up()

    def can_scroll_down(self):
        return self.can_scroll_page_down()

    def _scroll_position(self, delta):
        if self.is_enabled():
            model = self._slideable
            model.position = clamp(model.position + delta, 0, model.position_count - model.page_length)

    def _scroll_page(self, sign):
        if self.is_enabled():
            model = self._slideable
            remainder = (model.position - model.page_offset) % model.page_length
            delta = model.page_length - remainder if sign > 0 else -model.page_length if remainder == 0 else -remainder
            self._scroll_position(delta)

    def update(self):
        super().update()
        self._position_scroll.update()
        self._page_scroll.update()

    @listens("position")
    def __on_position_changed(self):
        self._position_scroll.update()
        self._page_scroll.update()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/accent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1183 bytes
from ...base import listenable_property
from ..component import Component
from ..control import ToggleButtonControl
from ..elements import NullFullVelocity


class AccentComponent(Component):
    accent_button = ToggleButtonControl(toggled_color="Accent.On", untoggled_color="Accent.Off")

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self.set_full_velocity(None)

    def set_full_velocity(self, full_velocity):
        self._full_velocity = full_velocity or NullFullVelocity()
        self.accent_button.is_toggled = self.activated

    @listenable_property
    def activated(self):
        return self._full_velocity.enabled

    @accent_button.toggled
    def accent_button(self, is_toggled, button):
        self._full_velocity.enabled = is_toggled
        self.notify_activated()

    @accent_button.released_delayed
    def accent_button(self, button):
        self.accent_button.is_toggled = False
        self._full_velocity.enabled = False
        self.notify_activated()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/session_navigation.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 6025 bytes
from ...base import listens
from ..component import Component
from .scroll import Scrollable, ScrollComponent


class SessionRingScroller(Scrollable):
    def __init__(self, session_ring=None, *a, **k):
        (super().__init__)(*a, **k)
        self._session_ring = session_ring

    do_scroll_up = NotImplemented
    do_scroll_down = NotImplemented

    def scroll_up(self):
        if self.can_scroll_up():
            self.do_scroll_up()

    def scroll_down(self):
        if self.can_scroll_down():
            self.do_scroll_down()


class SessionRingTrackScroller(SessionRingScroller):
    def can_scroll_up(self):
        return self._session_ring.track_offset > 0

    def can_scroll_down(self):
        return self._session_ring.track_offset + 1 < len(self._session_ring.tracks_to_use())

    def do_scroll_up(self):
        self._session_ring.move(-1, 0)

    def do_scroll_down(self):
        self._session_ring.move(1, 0)


class SessionRingSceneScroller(SessionRingScroller):
    def can_scroll_up(self):
        return self._session_ring.scene_offset > 0

    def can_scroll_down(self):
        return self._session_ring.scene_offset + 1 < len(self._session_ring.scenes())

    def do_scroll_up(self):
        self._session_ring.move(0, -1)

    def do_scroll_down(self):
        self._session_ring.move(0, 1)


class SessionRingTrackPager(SessionRingScroller):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self.page_size = self._session_ring.num_tracks

    def can_scroll_up(self):
        return self._session_ring.track_offset > 0

    def can_scroll_down(self):
        return self._session_ring.track_offset < len(self._session_ring.tracks_to_use()) - self.page_size

    def do_scroll_up(self):
        self._session_ring.set_offsets(
            max(0, self._session_ring.track_offset - self.page_size),
            self._session_ring.scene_offset,
        )

    def do_scroll_down(self):
        self._session_ring.set_offsets(
            self._session_ring.track_offset + self.page_size,
            self._session_ring.scene_offset,
        )


class SessionRingScenePager(SessionRingScroller):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self.page_size = self._session_ring.num_scenes

    def can_scroll_up(self):
        return self._session_ring.scene_offset > 0

    def can_scroll_down(self):
        return self._session_ring.scene_offset < len(self._session_ring.scenes()) - self.page_size

    def do_scroll_up(self):
        self._session_ring.set_offsets(
            self._session_ring.track_offset,
            max(0, self._session_ring.scene_offset - self.page_size),
        )

    def do_scroll_down(self):
        self._session_ring.set_offsets(
            self._session_ring.track_offset,
            self._session_ring.scene_offset + self.page_size,
        )


class SessionNavigationComponent(Component):
    track_scroller_type = SessionRingTrackScroller
    scene_scroller_type = SessionRingSceneScroller
    track_pager_type = SessionRingTrackPager
    scene_pager_type = SessionRingScenePager

    def __init__(self, session_ring=None, *a, **k):
        (super().__init__)(*a, **k)
        self._session_ring = session_ring
        self._SessionNavigationComponent__on_offset_changed.subject = self._session_ring
        self._SessionNavigationComponent__on_tracks_changed.subject = self._session_ring
        self._SessionNavigationComponent__on_scene_list_changed.subject = self.song
        self._vertical_banking = ScrollComponent((self.scene_scroller_type(session_ring)), parent=self)
        self._horizontal_banking = ScrollComponent((self.track_scroller_type(session_ring)), parent=self)
        self._vertical_paginator = ScrollComponent((self.scene_pager_type(session_ring)), parent=self)
        self._horizontal_paginator = ScrollComponent((self.track_pager_type(session_ring)), parent=self)

    @listens("offset")
    def __on_offset_changed(self, track_offset, _):
        self._update_vertical()
        self._update_horizontal()

    @listens("tracks")
    def __on_tracks_changed(self):
        self._update_horizontal()

    @listens("scenes")
    def __on_scene_list_changed(self):
        self._update_vertical()

    def _update_vertical(self):
        if self.is_enabled():
            self._vertical_banking.update()
            self._vertical_paginator.update()

    def _update_horizontal(self):
        if self.is_enabled():
            self._horizontal_banking.update()
            self._horizontal_paginator.update()

    def set_up_button(self, button):
        self._vertical_banking.set_scroll_up_button(button)

    def set_down_button(self, button):
        self._vertical_banking.set_scroll_down_button(button)

    def set_left_button(self, button):
        self._horizontal_banking.set_scroll_up_button(button)
        self._horizontal_banking.update()

    def set_right_button(self, button):
        self._horizontal_banking.set_scroll_down_button(button)

    def set_page_up_button(self, page_up_button):
        self._vertical_paginator.set_scroll_up_button(page_up_button)

    def set_page_down_button(self, page_down_button):
        self._vertical_paginator.set_scroll_down_button(page_down_button)

    def set_page_left_button(self, page_left_button):
        self._horizontal_paginator.set_scroll_up_button(page_left_button)

    def set_page_right_button(self, page_right_button):
        self._horizontal_paginator.set_scroll_down_button(page_right_button)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/controls/control_list.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3242 bytes
from ableton.v2.control_surface.control import ControlList as ControlListBase

from ...base import mixin
from . import RadioButtonGroup

_control_list_types = {}


def control_list(control_type, *a, **k):
    factory = _control_list_types.get(control_type)
    if not factory:
        factory = mixin(ControlList, control_type)
        _control_list_types[control_type] = factory
    return factory(control_type, *a, **k)


class ControlList(ControlListBase):
    class State(ControlListBase.State):
        def set_control_element_at_index(self, control_element, index):
            if self._control_elements:
                num_elements = len(self._control_elements)
                if num_elements > index:
                    self._control_elements[index] = control_element
                else:
                    self._control_elements.extend([None] * (index - num_elements) + [control_element])
            else:
                self._control_elements = [None] * index + [control_element]
            self._update_controls()


class FixedRadioButtonGroup(RadioButtonGroup):
    class State(RadioButtonGroup.State):
        def __init__(self, *a, **k):
            (super().__init__)(*a, **k)
            self._active_control_count = 0

        @property
        def active_control_count(self):
            return self._active_control_count

        @active_control_count.setter
        def active_control_count(self, control_count):
            self._active_control_count = control_count
            for index, control in enumerate(self._controls):
                control._get_state(self._manager).enabled = index < control_count
